import os
from pathlib import Path
from abc import ABCMeta, abstractmethod
import json
from json import JSONEncoder
import numpy as np
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge
from pyflowline.classes.flowline import pyflowline
from pyflowline.classes.confluence import pyconfluence
from pyflowline.formats.read_flowline import read_flowline_geojson
from pyflowline.formats.read_nhdplus_flowline_shapefile import read_nhdplus_flowline_shapefile_attribute
from pyflowline.formats.read_nhdplus_flowline_shapefile import extract_nhdplus_flowline_shapefile_by_attribute
from pyflowline.formats.read_nhdplus_flowline_shapefile import track_nhdplus_flowline
from pyflowline.formats.convert_shapefile_to_json import convert_shapefile_to_json
from pyflowline.formats.export_flowline import export_flowline_to_json
from pyflowline.formats.export_vertex import export_vertex_to_json
from pyflowline.algorithms.auxiliary.text_reader_string import text_reader_string
from pyflowline.algorithms.split.find_flowline_vertex import find_flowline_vertex
from pyflowline.algorithms.split.find_flowline_confluence import find_flowline_confluence
from pyflowline.algorithms.split.split_flowline import split_flowline
from pyflowline.algorithms.split.split_flowline_to_edge import split_flowline_to_edge
from pyflowline.algorithms.merge.merge_flowline import merge_flowline
from pyflowline.algorithms.direction.correct_flowline_direction import correct_flowline_direction
from pyflowline.algorithms.loop.remove_flowline_loop import remove_flowline_loop
from pyflowline.algorithms.simplification.remove_small_river import remove_small_river
from pyflowline.algorithms.simplification.remove_returning_flowline import remove_returning_flowline
from pyflowline.algorithms.simplification.remove_duplicate_flowline import remove_duplicate_flowline
from pyflowline.algorithms.index.define_stream_order import define_stream_order
from pyflowline.algorithms.index.define_stream_segment_index import define_stream_segment_index
from pyflowline.algorithms.intersect.intersect_flowline_with_mesh import intersect_flowline_with_mesh



class BasinClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson()) #lVertexID
        if isinstance(obj, pyedge):
            return obj.lEdgeID        
        if isinstance(obj, pyflowline):
            return obj.lFlowlineID
        if isinstance(obj, pyconfluence):
            return obj.dAngle_upstream
       
            
        return JSONEncoder.default(self, obj)



class pybasin(object):
    lBasinID =1 
    sBasinID=''
    lCellID_outlet=-1
    iFlag_debug = 0
    iFlag_disconnected =0
    iFlag_dam=0
    dLongitude_outlet_degree = -9999.
    dLatitude_outlet_degree = -9999.
    dAccumulation_threshold= 100000.0
    dThreshold_small_river = 10000
    dLength_flowline_filtered = 0.0
    dLength_flowline_simplified = 0.0
    dLength_flowline_conceptual = 0.0

    dArea_of_difference=0.0
    dDistance_displace = 0.0
    sWorkspace_output_basin=''
    sFilename_flowline_raw=''    
    sFilename_flowline_filter=''
    sFilename_flowline_filter_json=''
    sFilename_dam=''
    sFilename_flowline_topo=''
    #before intersect
    sFilename_flowline_simplified=''
    sFilename_flowline_segment_index_before_intersect=''
    sFilename_flowline_conceptual=''
    sFilename_flowline_edge=''
    sFilename_basin_info=''
    sFilename_flowline_simplified_info=''
    sFilename_flowline_conceptual_info=''
    sFilename_confluence_simplified_info=''
    sFilename_confluence_conceptual_info=''    
    aFlowline_basin_filtered=None
    aFlowline_basin_simplified=None
    aFlowline_basin_conceptual=None    
    pVertex_outlet=None
    aConfluence_basin_simplified= None
    aConfluence_basin_conceptual= None
    
    def __init__(self, aParameter):

        if 'lBasinID' in aParameter:            
            self.lBasinID             = int(aParameter['lBasinID'])
        else:
            self.lBasinID   = 1
        
        
        if 'lCellID_outlet' in aParameter:            
            self.lCellID_outlet             = int(aParameter['lCellID_outlet'])
        else:
            self.lCellID_outlet   = -1

        if 'iFlag_disconnected' in aParameter:            
            self.iFlag_disconnected             = int(aParameter['iFlag_disconnected'])
        else:
            self.iFlag_disconnected   = 0
        
        if 'iFlag_dam' in aParameter:            
            self.iFlag_dam             = int(aParameter['iFlag_dam'])
        else:
            self.iFlag_dam   = 0
        
        if 'iFlag_debug' in aParameter:            
            self.iFlag_debug             = int(aParameter['iFlag_debug'])
        else:
            self.iFlag_debug   = 0
        
        if 'dLongitude_outlet_degree' in aParameter:            
            self.dLongitude_outlet_degree             = float(aParameter['dLongitude_outlet_degree'])
        else:
            self.dLongitude_outlet_degree   = -9999.
        
        if 'dLatitude_outlet_degree' in aParameter:            
            self.dLatitude_outlet_degree             = float(aParameter['dLatitude_outlet_degree'])
        else:
            self.dLatitude_outlet_degree   = -9999.
        
        if 'dThreshold_small_river' in aParameter:            
            self.dThreshold_small_river             = float(aParameter['dThreshold_small_river'])
        else:
            self.dThreshold_small_river   = 10000.0

        if 'dAccumulation_threshold' in aParameter:            
            self.dAccumulation_threshold             = float(aParameter['dAccumulation_threshold'])
        else:
            self.dAccumulation_threshold = 100000.0   

        if 'sFilename_flowline_raw' in aParameter:
            self.sFilename_flowline_raw = aParameter['sFilename_flowline_raw']
        else:
            self.sFilename_flowline_raw   = ''
       
        if 'sFilename_flowline_filter' in aParameter:
            self.sFilename_flowline_filter = aParameter['sFilename_flowline_filter']
        else:
            self.sFilename_flowline_filter   = ''

        if 'sWorkspace_output_basin' in aParameter:
            self.sWorkspace_output_basin = aParameter['sWorkspace_output_basin']
        else:
            self.sWorkspace_output_basin   = '.'
            
        Path(self.sWorkspace_output_basin).mkdir(parents=True, exist_ok=True)

        self.sFilename_flowline_filter_json = os.path.join(str(self.sWorkspace_output_basin ), "flowline_filter.json"  )

        if 'sFilename_dam' in aParameter:
            self.sFilename_dam = aParameter['sFilename_dam']
        else:
            self.sFilename_dam   = ''

        if 'sFilename_flowline_topo' in aParameter:
            self.sFilename_flowline_topo = aParameter['sFilename_flowline_topo']
        else:
            self.sFilename_flowline_topo   =''

        self.sBasinID = sBasinID = "{:03d}".format(self.lBasinID)

        self.sFilename_flowline_segment_index_before_intersect = 'flowline_segment_index_before_intersect.json'
        self.sFilename_flowline_simplified = 'flowline_simplified.json'
        self.sFilename_flowline_intersect  = 'flowline_intersect_mesh.json'
        self.sFilename_flowline_conceptual = 'flowline_conceptual.json'
        self.sFilename_flowline_edge = 'flowline_edge.json'
        self.sFilename_area_of_difference = 'area_of_difference.json'
        self.sFilename_basin_info = 'basin_info.json'
        self.sFilename_flowline_conceptual_info = 'flowline_conceptual_info.json'
        self.sFilename_flowline_simplified_info = 'flowline_simplified_info.json'
        self.sFilename_confluence_conceptual_info = 'confluence_conceptual_info.json'
        self.sFilename_confluence_simplified_info = 'confluence_simplified_info.json'
        return
        
    def flowline_simplification(self):

        
        sFilename_flowline_filter = self.sFilename_flowline_filter
        sFilename_flowline_filter_json = self.sFilename_flowline_filter_json
        aFlowline_basin_filtered, pSpatial_reference = read_flowline_geojson( sFilename_flowline_filter_json )   
        sWorkspace_output_basin = self.sWorkspace_output_basin
        if self.iFlag_dam ==1:
            sFilename_dam = self.sFilename_dam
            aData_dam = text_reader_string(sFilename_dam, iSkipline_in =1,cDelimiter_in=',' )
            sFilename_flowline_topo = self.sFilename_flowline_topo
            aData_flowline_topo = text_reader_string(sFilename_flowline_topo, iSkipline_in =1,cDelimiter_in=',' )
            aFromFlowline = aData_flowline_topo[:,1].astype(int).ravel()
            aToFlowline = aData_flowline_topo[:,2].astype(int).ravel()
            sFilename_flowline_raw = self.sFilename_flowline_raw
            aNHDPlusID_filter = read_nhdplus_flowline_shapefile_attribute(sFilename_flowline_filter)
            aNHDPlusID_raw = read_nhdplus_flowline_shapefile_attribute(sFilename_flowline_raw)
            ndam = len(aData_dam)
            aNHDPlusID_dams_headwater = list()
            aNHDPlusID_dams_nonheadwater = list()
            for j in range(0, ndam):
                dLon = float(aData_dam[j][1])
                dLat = float(aData_dam[j][0])
                sDam = aData_dam[j][4]            
                lNHDPlusID = int(aData_dam[j][5])
                aNHDPlusID_dams_headwater.append(lNHDPlusID)
                if lNHDPlusID in aNHDPlusID_filter:
                    #remove by id
                    for k in range(len(aFlowline_basin_filtered)):
                        if aFlowline_basin_filtered[k].lNHDPlusID == lNHDPlusID:
                            aFlowline_basin_filtered.pop(k)
                            break
                    pass
                else:                                
                    aNHDPlusID_dam_nonheadwater = track_nhdplus_flowline(aNHDPlusID_filter, aFromFlowline, aToFlowline, lNHDPlusID)
                    aNHDPlusID_filter = aNHDPlusID_filter + aNHDPlusID_dams_headwater+ aNHDPlusID_dam_nonheadwater  
                    aNHDPlusID_dams_nonheadwater = aNHDPlusID_dams_nonheadwater + aNHDPlusID_dam_nonheadwater
            aFlowline_dams_headwater = extract_nhdplus_flowline_shapefile_by_attribute(sFilename_flowline_raw, aNHDPlusID_dams_headwater )
            for i in range(len(aFlowline_dams_headwater)):
                aFlowline_dams_headwater[i].iFlag_dam = 1
            aFlowline_dams_nonheadwater = extract_nhdplus_flowline_shapefile_by_attribute(sFilename_flowline_raw, aNHDPlusID_dams_nonheadwater )
            aFlowline_basin_filtered = aFlowline_basin_filtered + aFlowline_dams_headwater + aFlowline_dams_nonheadwater
        else:
            pass
        if self.iFlag_disconnected == 1:
            #not used anymore                
            #aThreshold = np.full(2, 300.0, dtype=float)
            #aFlowline_basin_filtered = connect_disconnect_flowline(aFlowline_basin_filtered, aVertex, aThreshold)
            #sFilename_out = 'flowline_connect.json'
            #sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)    
            #export_flowline_to_json(iFlag_projected, aFlowline_basin_filtered,pSpatial_reference_gcs, sFilename_out)
            pass
        else:
            pass
            
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_before_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)            
            self.export_flowline(aFlowline_basin_filtered, sFilename_out)
        #calculate length
        self.aFlowline_basin_filtered = aFlowline_basin_filtered
        self.dLength_flowline_filtered = self.calculate_flowline_length(aFlowline_basin_filtered)

        #simplification started
        aVertex = find_flowline_vertex(aFlowline_basin_filtered)
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_vertex_without_confluence_before_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
            export_vertex_to_json( aVertex, sFilename_out)
        aFlowline_basin_simplified = split_flowline(aFlowline_basin_filtered, aVertex)
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_split_by_point_before_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
            export_flowline_to_json(aFlowline_basin_simplified, sFilename_out)
        #ues location to find outlet
        point= dict()   
        point['dLongitude_degree'] = self.dLongitude_outlet_degree
        point['dLatitude_degree'] = self.dLatitude_outlet_degree
        pVertex_outlet=pyvertex(point)
        aFlowline_basin_simplified = correct_flowline_direction(aFlowline_basin_simplified,  pVertex_outlet )
        pVertex_outlet = aFlowline_basin_simplified[0].pVertex_end
        self.pVertex_outlet = pVertex_outlet
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_direction_before_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
            export_flowline_to_json( aFlowline_basin_simplified,  sFilename_out)
        #step 4: remove loops
        aFlowline_basin_simplified = remove_flowline_loop(aFlowline_basin_simplified)    
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_loop_before_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
            export_flowline_to_json( aFlowline_basin_simplified, sFilename_out)
        #using loop to remove small river, here we use 5 steps
        for i in range(3):
            sStep = "{:02d}".format(i+1)
            aFlowline_basin_simplified = remove_small_river(aFlowline_basin_simplified, self.dThreshold_small_river)
            if self.iFlag_debug ==1:
                sFilename_out = 'flowline_large_'+ sStep +'_before_intersect.json'
                sFilename_out =os.path.join(sWorkspace_output_basin, sFilename_out)
                export_flowline_to_json( aFlowline_basin_simplified,  sFilename_out)
            aVertex, lIndex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence, aConnectivity = find_flowline_confluence(aFlowline_basin_simplified,  pVertex_outlet)
            if self.iFlag_debug ==1:
                sFilename_out = 'flowline_vertex_with_confluence_'+ sStep +'_before_intersect.json'
                sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
                export_vertex_to_json( aVertex,  sFilename_out, aAttribute_data=aConnectivity)
            aFlowline_basin_simplified = merge_flowline( aFlowline_basin_simplified,aVertex, pVertex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence  )  
            if self.iFlag_debug ==1:
                sFilename_out = 'flowline_merge_'+ sStep +'_before_intersect.json'
                sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
                export_flowline_to_json( aFlowline_basin_simplified,  sFilename_out)
            if len(aFlowline_basin_simplified) == 1:
                break
        
        #the final vertex info
        aVertex, lIndex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence, aConnectivity = find_flowline_confluence(aFlowline_basin_simplified,  pVertex_outlet)
        sFilename_out = 'vertex_simplified.json'
        sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
        export_vertex_to_json( aVertex,  sFilename_out, aAttribute_data=aConnectivity)
        
        #self.dLength_flowline_simplified = self.calculate_flowline_length(aFlowline_basin_simplified)
        aVertex = np.array(aVertex)
        aIndex_confluence = np.array(aIndex_confluence)
        if aIndex_confluence.size > 0:        
            aVertex_confluence = aVertex[aIndex_confluence]
            self.aConfluence_basin_simplified = self.build_confluence(aFlowline_basin_simplified, aVertex_confluence) 
        

        #build segment index
        aFlowline_basin_simplified, aStream_segment = define_stream_segment_index(aFlowline_basin_simplified)
        if self.iFlag_debug ==1:
            sFilename_out = self.sFilename_flowline_segment_index_before_intersect
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
            export_flowline_to_json(  aFlowline_basin_simplified, sFilename_out, \
                aAttribute_data=[aStream_segment], aAttribute_field=['iseg'], aAttribute_dtype=['int'])
        #build stream order 
        aFlowline_basin_simplified, aStream_order = define_stream_order(aFlowline_basin_simplified)
        sFilename_out = self.sFilename_flowline_simplified
        sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
        export_flowline_to_json(  aFlowline_basin_simplified, sFilename_out, \
                aAttribute_data=[aStream_segment, aStream_order], aAttribute_field=['iseg','iord'], aAttribute_dtype=['int','int'])
        

        self.aFlowline_basin_simplified= aFlowline_basin_simplified
        return aFlowline_basin_simplified

    def reconstruct_topological_relationship(self, iMesh_type, sFilename_mesh):
        
        sWorkspace_output_basin = self.sWorkspace_output_basin
        sFilename_flowline = self.sFilename_flowline_simplified
        sFilename_flowline_in = os.path.join(sWorkspace_output_basin, sFilename_flowline)
        aFlowline_basin_simplified, pSpatial_reference = read_flowline_geojson( sFilename_flowline_in )   
                
        sFilename_flowline_intersect = self.sFilename_flowline_intersect
        sFilename_flowline_intersect_out = os.path.join(sWorkspace_output_basin, sFilename_flowline_intersect)
        aCell, aCell_intersect_basin, aFlowline_intersect_all = intersect_flowline_with_mesh(iMesh_type, sFilename_mesh, \
            sFilename_flowline_in, sFilename_flowline_intersect_out)
        sFilename_flowline_filter_json = self.sFilename_flowline_filter
        
        point= dict()
        point['dLongitude_degree'] = self.dLongitude_outlet_degree
        point['dLatitude_degree'] = self.dLatitude_outlet_degree
        pVertex_outlet_initial=pyvertex(point)

        #from this point, aFlowline_basin is conceptual
        #segment based
        aFlowline_basin_conceptual, lCellID_outlet, pVertex_outlet = remove_returning_flowline(iMesh_type, aCell_intersect_basin, pVertex_outlet_initial)
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_simplified_after_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)  
            export_flowline_to_json(aFlowline_basin_conceptual,  sFilename_out)

        #edge based
        aFlowline_basin_conceptual, aEdge = split_flowline_to_edge(aFlowline_basin_conceptual)
        aFlowline_basin_conceptual = remove_duplicate_flowline(aFlowline_basin_conceptual)
        aFlowline_basin_conceptual = correct_flowline_direction(aFlowline_basin_conceptual,  pVertex_outlet )
        aFlowline_basin_conceptual = remove_flowline_loop(  aFlowline_basin_conceptual )  
  
        aVertex, lIndex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence, aConnectivity\
            = find_flowline_confluence(aFlowline_basin_conceptual,  pVertex_outlet)
        if self.iFlag_debug ==1:
            sFilename_out = 'flowline_vertex_with_confluence_after_intersect.json'
            sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
            export_vertex_to_json( aVertex,  sFilename_out, aAttribute_data=aConnectivity)
        
        #segment based
        aFlowline_basin_conceptual = merge_flowline( aFlowline_basin_conceptual,aVertex, pVertex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence  )                          
        aVertex, lIndex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence, aConnectivity\
            = find_flowline_confluence(aFlowline_basin_conceptual,  pVertex_outlet)        
        aFlowline_basin_conceptual, aStream_segment = define_stream_segment_index(aFlowline_basin_conceptual)
        aFlowline_basin_conceptual, aStream_order = define_stream_order(aFlowline_basin_conceptual)

        #save confluence
        aVertex = np.array(aVertex)
        aIndex_confluence = np.array(aIndex_confluence)
        if aIndex_confluence.size > 0:        
            aVertex_confluence = aVertex[aIndex_confluence] 
            self.aConfluence_basin_conceptual = self.build_confluence(aFlowline_basin_conceptual, aVertex_confluence) 
          

        #edge based
        aFlowline_basin_edge, aEdge = split_flowline_to_edge(aFlowline_basin_conceptual)
        sFilename_out = self.sFilename_flowline_edge
        sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
        export_flowline_to_json(  aFlowline_basin_edge, sFilename_out)

        sFilename_out = self.sFilename_flowline_conceptual
        sFilename_out = os.path.join(sWorkspace_output_basin, sFilename_out)
        export_flowline_to_json(  aFlowline_basin_conceptual, sFilename_out, \
            aAttribute_data=[aStream_segment, aStream_order], aAttribute_field=['iseg','iord'], aAttribute_dtype=['int','int'])

        self.aFlowline_basin_conceptual = aFlowline_basin_conceptual     
        
        self.lCellID_outlet = lCellID_outlet
        self.dLongitude_outlet_degree = pVertex_outlet.dLongitude_degree
        self.dLatitude_outlet_degree = pVertex_outlet.dLatitude_degree
        

        return aCell_intersect_basin

    def build_confluence(self, aFlowline_basin_in, aVertex_confluence_in):    
        #this can only be calculated for confluence
        aConfluence_basin=list()
        for pVertex in aVertex_confluence_in:   
            aFlowline_upstream =list()
            for pFlowline in aFlowline_basin_in:
                pVertex_start = pFlowline.pVertex_start
                pVertex_end = pFlowline.pVertex_end
                if pVertex_end == pVertex:                 
                    aFlowline_upstream.append(pFlowline)
                    pass
                if pVertex_start == pVertex:
                    pFlowline_downstream=pFlowline

            pConfluence = pyconfluence(pVertex, aFlowline_upstream, pFlowline_downstream)
            aConfluence_basin.append(pConfluence)   
        return aConfluence_basin

    def analyze(self):      
        if self.aFlowline_basin_filtered is None:
            sFilename_flowline_filter = self.sFilename_flowline_filter
            sFilename_flowline_filter_json = self.sFilename_flowline_filter_json
            self.aFlowline_basin_filtered, pSpatial_reference = read_flowline_geojson( sFilename_flowline_filter_json )   
            self.dLength_flowline_filtered = self.calculate_flowline_length(self.aFlowline_basin_filtered)
        
        point= dict()
        point['dLongitude_degree'] = self.dLongitude_outlet_degree
        point['dLatitude_degree'] = self.dLatitude_outlet_degree
        pVertex_outlet_initial=pyvertex(point)
        if self.aFlowline_basin_simplified is None:
            sFilename_flowline = self.sFilename_flowline_simplified
            sFilename_flowline_in = os.path.join(self.sWorkspace_output_basin, sFilename_flowline)
            aFlowline_simplified,pSpatial_reference = read_flowline_geojson( sFilename_flowline_in )   
            read_flowline_geojson
            self.aFlowline_basin_simplified = aFlowline_simplified
            aVertex, lIndex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence, aConnectivity\
            = find_flowline_confluence(self.aFlowline_basin_simplified,  pVertex_outlet_initial)  
            aVertex = np.array(aVertex)
            aIndex_confluence = np.array(aIndex_confluence)
            if aIndex_confluence.size > 0:        
                aVertex_confluence = aVertex[aIndex_confluence] 
                self.aConfluence_basin_simplified = self.build_confluence(self.aFlowline_basin_simplified, aVertex_confluence)

        self.dLength_flowline_simplified = self.calculate_flowline_length(self.aFlowline_basin_simplified)

        if self.aFlowline_basin_conceptual is None:
            sFilename_flowline = self.sFilename_flowline_conceptual
            sFilename_flowline_in = os.path.join(self.sWorkspace_output_basin, sFilename_flowline)
            aFlowline_conceptual, pSpatial_reference = read_flowline_geojson( sFilename_flowline_in )   
            self.aFlowline_basin_conceptual = aFlowline_conceptual
            
            aVertex, lIndex_outlet, aIndex_headwater,aIndex_middle, aIndex_confluence, aConnectivity\
            = find_flowline_confluence(self.aFlowline_basin_conceptual,  pVertex_outlet_initial)  
            aVertex = np.array(aVertex)
            aIndex_confluence = np.array(aIndex_confluence)
            if aIndex_confluence.size > 0:        
                aVertex_confluence = aVertex[aIndex_confluence] 
                self.aConfluence_basin_conceptual = self.build_confluence(self.aFlowline_basin_conceptual, aVertex_confluence)

        self.dLength_flowline_conceptual = self.calculate_flowline_length(self.aFlowline_basin_conceptual)
        self.calculate_river_sinuosity()
        self.calculate_confluence_branching_angle()
        return    
    
    def export(self):
        self.export_basin_info_to_json()
        self.export_flowline_info_to_json()
        self.export_confluence_info_to_json()        
        return

    def export_flowline(self, aFlowline_in, sFilename_json_in,iFlag_projected_in = None,  pSpatial_reference_in = None):
        export_flowline_to_json(aFlowline_in, sFilename_json_in,\
            iFlag_projected_in= iFlag_projected_in, \
            pSpatial_reference_in = pSpatial_reference_in)

    def export_basin_info_to_json(self):
        sFilename_json = self.sFilename_basin_info
        sFilename_json = os.path.join(str(Path(self.sWorkspace_output_basin)  ) , sFilename_json  )

        aSkip = ['aFlowline_basin_filtered', \
                'aFlowline_basin_simplified','aFlowline_basin_conceptual','aConfluence_basin_simplified',
                'aConfluence_basin_conceptual']
        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        with open(sFilename_json, 'w', encoding='utf-8') as f:
            sJson = json.dumps(obj, default=lambda o: o.__dict__,\
            sort_keys=True, \
                indent = 4, \
                    ensure_ascii=True, \
                        cls=BasinClassEncoder)      
            f.write(sJson)    
            f.close()
        return

    def export_flowline_info_to_json(self):
        iFlag_export_simplified=0
        if iFlag_export_simplified==1:
            sFilename_json = self.sFilename_flowline_simplified_info
            sFilename_json = os.path.join(str(Path(self.sWorkspace_output_basin)  ) , sFilename_json  )
            with open(sFilename_json, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aFlowline_basin_simplified], indent = 4)        
                f.write(sJson)    
                f.close()

        sFilename_json = self.sFilename_flowline_conceptual_info
        sFilename_json = os.path.join(str(Path(self.sWorkspace_output_basin)  ) , sFilename_json  )

        
        with open(sFilename_json, 'w', encoding='utf-8') as f:
            sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aFlowline_basin_conceptual], indent = 4)        
            f.write(sJson)    
            f.close()
        return

    def export_confluence_info_to_json(self):
        iFlag_export_confluence =0
        if iFlag_export_confluence==1:
            sFilename_json = self.sFilename_confluence_simplified_info
            sFilename_json = os.path.join(str(Path(self.sWorkspace_output_basin)  ) , sFilename_json  )
            with open(sFilename_json, 'w', encoding='utf-8') as f:
                sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aConfluence_basin_simplified], indent = 4)        
                f.write(sJson)    
                f.close()

        sFilename_json = self.sFilename_confluence_conceptual_info
        sFilename_json = os.path.join(str(Path(self.sWorkspace_output_basin)  ) , sFilename_json  )
        
        with open(sFilename_json, 'w', encoding='utf-8') as f:
            sJson = json.dumps([json.loads(ob.tojson()) for ob in self.aConfluence_basin_conceptual], indent = 4)        
            f.write(sJson)    
            f.close()
        return   

    def tojson(self):
        aSkip = ['aFlowline_basin_filtered', \
                'aFlowline_basin_simplified','aFlowline_basin_conceptual','aConfluence_basin_simplified',
                'aConfluence_basin_conceptual']

        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
 
    
        sJson = json.dumps(obj, \
            sort_keys=True, \
                indent = 4, \
                    ensure_ascii=True, \
                        cls=BasinClassEncoder)
        return sJson
    
    def export_config_to_json(self, sFilename_output_in = None):
        #single basin
        if sFilename_output_in is not None:
            sFilename_output = sFilename_output_in
        else:
            sFilename_output = os.path.join(self.sWorkspace_output_basin, 'configuration_basin.json' )

        aSkip = ['aFlowline_basin_filtered', \
                'aFlowline_basin_simplified','aFlowline_basin_conceptual','aConfluence_basin_simplified',
                'aConfluence_basin_conceptual']
        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        with open(sFilename_output, 'w', encoding='utf-8') as f:
            json.dump(obj, f,sort_keys=True, \
                ensure_ascii=False, \
                indent=4, \
                cls=BasinClassEncoder)
        return

    def convert_flowline_to_json(self):
        sFilename_raw = self.sFilename_flowline_filter            
        sFilename_out = self.sFilename_flowline_filter_json
        print('This is the filtered flowline:', sFilename_raw )
        convert_shapefile_to_json(1, sFilename_raw, sFilename_out)
        
    def calculate_flowline_length(self, aFlowline_in):
        dLength = 0.0
        nflowline = len(aFlowline_in)
        for i in range(nflowline):
            pFlowline= aFlowline_in[i]
            pFlowline.calculate_length()
            dLength = dLength + pFlowline.dLength        
        return dLength

    def calculate_river_sinuosity(self):
        for pFlowline in self.aFlowline_basin_simplified:
            pFlowline.calculate_flowline_sinuosity() 

        for pFlowline in self.aFlowline_basin_conceptual:
            pFlowline.calculate_flowline_sinuosity()    

        return

    def calculate_confluence_branching_angle(self):
        for pConfluence in self.aConfluence_basin_simplified:
            pConfluence.calculate_branching_angle()
        for pConfluence in self.aConfluence_basin_conceptual:
            pConfluence.calculate_branching_angle()    
        return
    


import os, sys

import numpy as np


from osgeo import ogr
from pyflowline.classes.tin import pytin

from pyflowline.formats.convert_coordinates import convert_pcs_coordinates_to_cell

def create_tin_mesh(dX_left_in, dY_bot_in, dResolution_meter_in, ncolumn_in, nrow_in,
sFilename_output_in, sFilename_spatial_reference_in):
     
    
    if os.path.exists(sFilename_output_in): 
        #delete it if it exists
        os.remove(sFilename_output_in)
    
    pDriver_shapefile = ogr.GetDriverByName('Esri Shapefile')

    
    pDataset = pDriver_shapefile.CreateDataSource(sFilename_output_in)
    
    pDataset_shapefile = pDriver_shapefile.Open(sFilename_spatial_reference_in, 0)
    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
    pSrs = pLayer_shapefile.GetSpatialRef()

    #pSrs = osr.SpatialReference()  
    #pSrs.ImportFromEPSG(4326)    # WGS84 lat/lon

    pLayer = pDataset.CreateLayer('cell', pSrs, ogr.wkbPolygon)
    # Add one attribute
    pLayer.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    
    pLayerDefn = pLayer.GetLayerDefn()
    pFeature = ogr.Feature(pLayerDefn)

    

    xleft = dX_left_in
    ybottom = dY_bot_in

    dArea = np.power(dResolution_meter_in,2.0)
    #tin edge
    dLength_edge = np.sqrt(  4.0 * dArea /  np.sqrt(3.0) )  
    dX_shift = 0.5 * dLength_edge
    dY_shift = 0.5 * dLength_edge * np.sqrt(3.0) 
    dX_spacing = dX_shift * 2
    dY_spacing = dY_shift

    lID =0 

    #geojson
    aTin=list()
    #.........
    #(x2,y2)-----(x3,y3)
    #   |           |
    #(x1,y1)-----(x4,y4)
    #...............
    for column in range(0, ncolumn_in):
  
        for row in range(0, nrow_in):
            

            if column % 2 == 0 :
                if row % 2 == 0:
                    #define a polygon here
                    x1 = xleft + (column * dX_shift)
                    y1 = ybottom + (row * dY_spacing)
                    x2 = x1 + dX_spacing
                    y2 = y1 
                    x3 = x1 + dX_shift
                    y3 = y1 + dY_spacing
                else:
                    x1 = xleft + (column * dX_shift) 
                    y1 = ybottom + (row +1)* dY_spacing 
                    x2 = x1 + dX_shift
                    y2 = y1 - dY_shift    
                    x3 = x1 + dX_spacing
                    y3 = y1  
                    
            else:
                if row % 2 == 0:
                    x1 = xleft + column *  dX_shift
                    y1 = ybottom + (row + 1)* dY_spacing
                    x2 = x1 + dX_shift
                    y2 = y1 - dY_shift    
                    x3 = x1 + dX_spacing
                    y3 = y1   
                else:
                    x1 = xleft + column *  dX_shift
                    y1 = ybottom + (row )* dY_spacing
                    x2 = x1 + dX_spacing
                    y2 = y1   
                    x3 = x1 + dX_shift
                    y3 = y1 + dY_spacing
                         

                   
           
            aCoords = np.full((4,2), -9999.0, dtype=float)

            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(x1, y1)
            ring.AddPoint(x2, y2)
            ring.AddPoint(x3, y3)
            
            ring.AddPoint(x1, y1)
            pPolygon = ogr.Geometry(ogr.wkbPolygon)
            pPolygon.AddGeometry(ring)

            pFeature.SetGeometry(pPolygon)
            pFeature.SetField("id", lID)
            pLayer.CreateFeature(pFeature)

            lID = lID + 1

            #dummy = loads( ring.ExportToWkt() )
            #aCoords = dummy.exterior.coords
            aCoords[0,0] = x1
            aCoords[0,1] = y1
            aCoords[1,0] = x2
            aCoords[1,1] = y2
            aCoords[2,0] = x3
            aCoords[2,1] = y3
            
            aCoords[3,0] = x1
            aCoords[3,1] = y1
            
            dummy1= np.array(aCoords)
            pHexagon = convert_pcs_coordinates_to_cell(1, dummy1)
            aTin.append(pHexagon)

            pass
        
    pDataset = pLayer = pFeature  = None      


    return aTin



import os,sys
from math import cos, sin, sqrt, acos
import numpy as np
import osgeo
from osgeo import  osr, gdal
#most of these functions are copied from the pyearth package

#https://stackoverflow.com/questions/8204998/how-to-check-if-a-pointlonc-latc-lie-on-a-great-circle-running-from-lona-lata
def calculate_distance_to_plane(x1, y1, x2, y2, x3, y3):
    
    a = np.radians(np.array((x1, y1) ))
    b = np.radians(np.array((x2, y2) )) #this is the middle one
    c = np.radians(np.array((x3, y3) ))
    # The points in 3D space
    a3 = longlat_to_3d(*a)
    b3 = longlat_to_3d(*b)
    c3 = longlat_to_3d(*c)
    #The formula is x+b*y+c*z=0 
    x1,y1,z1 = a3
    x2,y2,z2 = b3
    x3,y3,z3 = c3
    c = (-x1*y3 + x3* y1)/( z1*y3 - z3*y1 )
    b = (  -x1*z3 + x3 * z1 ) / (y1 * z3 - y3*z1)
    distance = abs(  x2 + b * y2 + c * z2 )
    return distance

def calculate_angle_betwen_vertex_normal(x1, y1, x2, y2, x3, y3):
    a = np.radians(np.array((x1, y1) ))
    b = np.radians(np.array((x2, y2) ))
    c = np.radians(np.array((x3, y3) ))
    # The points in 3D space
    a3 = longlat_to_3d(*a)
    b3 = longlat_to_3d(*b)
    c3 = longlat_to_3d(*c)
    a3vec = a3 - b3
    c3vec = c3 - b3 
    dot=np.dot(a3vec, c3vec)
    g = np.cross(a3vec, c3vec)
    det = np.dot(b3, g)
    angle = np.arctan2(det, dot)
    f = np.degrees(angle) 
    if f < 0:
        f = 360 + f
    
    return f

def calculate_angle_betwen_vertex(x1, y1, x2, y2, x3, y3):
    #all in degree
    a = np.radians(np.array((x1, y1) ))
    b = np.radians(np.array((x2, y2) ))
    c = np.radians(np.array((x3, y3) ))
    # The points in 3D space
    a3 = longlat_to_3d(*a)
    b3 = longlat_to_3d(*b)
    c3 = longlat_to_3d(*c)
    # Vectors in 3D space
    a3vec = a3 - b3
    c3vec = c3 - b3        
    angle3deg = angle_between_vectors_degrees(a3vec, c3vec)
    return  angle3deg
    
def longlat_to_3d(lonr, latr):
    """Convert a point given latitude and longitude in radians to
    3-dimensional space, assuming a sphere radius of one."""
    a = cos(latr) * cos(lonr)
    b = cos(latr) * sin(lonr)
    c = sin(latr)
    return np.array((a,b,c))

def angle_between_vectors_degrees(u, v):
    """Return the angle between two vectors in any dimension space,
    in degrees.
    """
    a = np.dot(u, v)
    b = np.linalg.norm(u)
    c = np.linalg.norm(v)
    d = a / (b* c)
    if d > 1:
        d = 1
    if d < -1:
        d = -1
    e = acos(d)
    f = np.degrees(e)
    return f
    

def convert_360_to_180(dLongitude_in):
    """[This function is modified from
    http://www.idlcoyote.com/map_tips/lonconvert.html]

    Args:
        dLongitude_in ([type]): [description]

    Returns:
        [type]: [description]
    """ 
    a = int(dLongitude_in /180)
    dLongitude_out = dLongitude_in - a*360.0
    return dLongitude_out

def convert_180_to_360(dLongitude_in):
    """[This function is modified from
    http://www.idlcoyote.com/map_tips/lonconvert.html]

    Args:
        dLongitude_in ([type]): [description]

    Returns:
        [type]: [description]
    """

    dLongitude_out = (dLongitude_in + 360.0) % 360.0

    return dLongitude_out

def retrieve_geotiff_metadata(sFilename_geotiff_in):
    """[retrieve the metadata of a geotiff file]

    Args:
        sFilename_geotiff_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    pDriver = gdal.GetDriverByName('GTiff')
   
    pDataset = gdal.Open(sFilename_geotiff_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_geotiff_in)
        sys.exit("Try again!")
    else: 
        pProjection = pDataset.GetProjection()
        pSpatial_reference = osr.SpatialReference(wkt=pProjection)    
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize        
        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        dPixelWidth = pGeotransform[1]
        pPixelHeight = pGeotransform[5]               
        return dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, pSpatial_reference, pProjection, pGeotransform


def degree_to_meter(dLatitude_in, dResolution_degree_in):
    """[Conver a degree-based resolution to meter-based resolution]

    Args:
        dLatitude_in ([type]): [description]
        dResolution_degree_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    dRadius = 6378137.0  #unit: m earth radius
    dRadius2 = dRadius * np.cos( dLatitude_in / 180.0 * np.pi)
    dResolution_meter = dResolution_degree_in / 360.0 * (2*np.pi * dRadius2)

    return dResolution_meter

def meter_to_degree(dResolution_meter_in, dLatitude_mean_in):
    """[Convert a meter-based resolution to degree-based resolution]

    Args:
        dResolution_meter_in ([type]): [description]
        dLatitude_mean_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    dLatitude_mean = abs(dLatitude_mean_in)

    dRadius = 6378137.0 # #unit: m earth radius
    dRadius2 = dRadius * np.cos( dLatitude_mean / 180.0 * np.pi)

    ##dResolution_meter = dResolution_degree / 360.0 * 2*np.pi * dRadius2

    dResolution_degree= dResolution_meter_in/(2*np.pi * dRadius2) * 360.0

    return dResolution_degree



def reproject_coordinates(x_in, y_in, spatial_reference_source, spatial_reference_target=None):
    """[Reproject coordinates from one reference to another. By default to WGS84.]

    Args:
        x_in ([type]): [description]
        y_in ([type]): [description]
        spatial_reference_source ([type]): [description]
        spatial_reference_target ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """      
    if spatial_reference_target is not None:
        pass
    else:
        spatial_reference_target = osr.SpatialReference()
        spatial_reference_target.ImportFromEPSG(4326)        
        pass
    
    if int(osgeo.__version__[0]) >= 3:
    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
                    
        spatial_reference_source.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        spatial_reference_target.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

    
    pTransform = osr.CoordinateTransformation( spatial_reference_source, spatial_reference_target)
   
    x_new,y_new, z = pTransform.TransformPoint( x_in,y_in)
    
    return x_new,y_new

def reproject_coordinates_batch(aX_in, aY_in, spatial_reference_source, spatial_reference_target=None):
    """[Reproject coordinates from one reference to another in batch mode. By default to WGS84.]

    Args:
        aX_in ([type]): [description]
        aY_in ([type]): [description]
        spatial_reference_source ([type]): [description]
        spatial_reference_target ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """
    #Reproject a list of x,y coordinates. 

    if spatial_reference_target is not None:

        pass
    else:
        spatial_reference_target = osr.SpatialReference()
        spatial_reference_target.ImportFromEPSG(4326)
        
        pass

    
    if int(osgeo.__version__[0]) >= 3:
    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
                    
        spatial_reference_source.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
        spatial_reference_target.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

    
    pTransform = osr.CoordinateTransformation( spatial_reference_source, spatial_reference_target)

    npoint = len(aX_in)
    x_new=list()
    y_new=list()
    for i in range(npoint):
        x0 = aX_in[i]
        y0 = aY_in[i]
   
        x1,y1, z = pTransform.TransformPoint( x0,y0)

        x_new.append(x1)
        y_new.append(y1)
    
    return x_new,y_new


def calculate_distance_based_on_lon_lat(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points 
    on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians 
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    # haversine formula 
    dlon = lon2 - lon1 
    dlat = lat2 - lat1 
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * np.arcsin(sqrt(a)) 
    # Radius of earth in kilometers. Use 3956 for miles
    r = 6378137.0
    return c * r

def calculate_polygon_area(lons, lats,  algorithm = 0, radius = 6378137.0):
    """
    Computes area of spherical polygon, assuming spherical Earth. 
    Returns result in ratio of the sphere's area if the radius is specified. Otherwise, in the units of provided radius.
    lats and lons are in degrees.
    """
    #TODO: take into account geodesy (i.e. convert latitude to authalic sphere, use radius of authalic sphere instead of mean radius of spherical earth)
    lats = np.deg2rad(lats)
    lons = np.deg2rad(lons)

    if algorithm==0:
        # Line integral based on Green's Theorem, assumes spherical Earth      
        #close polygon
        if lats[0]!=lats[-1]:
            lats = np.append(lats, lats[0])
            lons = np.append(lons, lons[0])

        # Get colatitude (a measure of surface distance as an angle)
        a = np.sin(lats/2)**2 + np.cos(lats)* np.sin(lons/2)**2
        colat = 2*np.arctan2( np.sqrt(a), np.sqrt(1-a) )

        #azimuth of each point in segment from the arbitrary origin
        az = np.arctan2(np.cos(lats) * np.sin(lons), np.sin(lats)) % (2*np.pi)

        # Calculate step sizes
        # daz = np.diff(az) % (2*pi)
        daz = np.diff(az)
        daz = (daz + np.pi) % (2 * np.pi) - np.pi

        # Determine average surface distance for each step
        deltas=np.diff(colat)/2
        colat=colat[0:-1]+deltas

        # Integral over azimuth is 1-cos(colatitudes)
        integrands = (1-np.cos(colat)) * daz

        # Integrate and save the answer as a fraction of the unit sphere.
        # Note that the sum of the integrands will include a factor of 4pi.
        area = abs(sum(integrands))/(4*np.pi) # Could be area of inside or outside

        area = min(area,1-area)
        if radius is not None: #return in units of radius
            return area * 4*np.pi*radius**2
        else: #return in ratio of sphere total area
            return area
    elif algorithm==2:
        #L'Huilier Theorem, assumes spherical earth
        #see:
        # https://mathworld.wolfram.com/SphericalPolygon.html
        # https://web.archive.org/web/20160324191929/http://forum.worldwindcentral.com/showthread.php?20724-A-method-to-compute-the-area-of-a-spherical-polygon
        # https://github.com/spacetelescope/spherical_geometry/blob/master/spherical_geometry/polygon.py
        # https://github.com/tylerjereddy/spherical-SA-docker-demo/blob/master/docker_build/demonstration.py
        #TODO
        pass
    elif algorithm==3:
        #https://trs.jpl.nasa.gov/handle/2014/41271
        #TODO
        pass


def gdal_read_geotiff_file(sFilename_in):
    """Read a Geotiff format raster file.

    Args:
        sFilename_in (string): The file name

    Returns:
        tuple: aData_out, pPixelWidth, dOriginX, dOriginY, nrow, ncolumn, dMissing_value , pGeotransform, pProjection,  pSpatial_reference
    """
    
    if os.path.exists(sFilename_in):
        pass
    else:
        print('The file does not exist!')
        return

    sDriverName='GTiff'
    pDriver = gdal.GetDriverByName(sDriverName)  

    if pDriver is None:
        print ("%s pDriver not available.\n" % sDriverName)
    else:
        print  ("%s pDriver IS available.\n" % sDriverName)  

    pDataset = gdal.Open(sFilename_in, gdal.GA_ReadOnly)

    if pDataset is None:
        print("Couldn't open this file: " + sFilename_in)
        sys.exit("Try again!")
    else:       
        pProjection = pDataset.GetProjection()

        pDataset.GetMetadata()
       
        ncolumn = pDataset.RasterXSize
        nrow = pDataset.RasterYSize
        nband = pDataset.RasterCount

        pGeotransform = pDataset.GetGeoTransform()
        dOriginX = pGeotransform[0]
        dOriginY = pGeotransform[3]
        dPixelWidth = pGeotransform[1]
        pPixelHeight = pGeotransform[5]

        pBand = pDataset.GetRasterBand(1)

        # Data type of the values
        gdal.GetDataTypeName(pBand.DataType)
        # Compute statistics if needed
        if pBand.GetMinimum() is None or pBand.GetMaximum() is None:
            pBand.ComputeStatistics(0)

        dMissing_value = pBand.GetNoDataValue()
       
        aData_out = pBand.ReadAsArray(0, 0, ncolumn, nrow)
    
        #we will use one of them to keep the consistency
        pSpatial_reference = osr.SpatialReference(wkt=pProjection)
       

        pDataset = None
        pBand = None      
        pBand = None

        return aData_out, dPixelWidth, dOriginX, dOriginY, nrow, ncolumn, dMissing_value, pGeotransform, pProjection,  pSpatial_reference

def Google_MetersPerPixel( zoomLevel ):  
   
   # Return to the caller if there is an error.
   #On_Error, 2
   
   #; Need a zoom level?
   
   
   # Number of pixels in an image with a zoom level of 0.
   pixels_in_image = 256
   
   # The equitorial radius of the Earth assuming WGS-84 ellipsoid.
   earth_radius = 6378137.0
   
   # The number of meters per pixel.
   metersPerPixel = (2* np.pi * earth_radius) / pixels_in_image / np.power(2,zoomLevel)
   
   # Return the value.
   return metersPerPixel
   





import os
import numpy as np
from osgeo import ogr, osr
from shapely.wkt import loads

def intersect_flowline_with_vertex( sFilename_flowline_in, sFilename_vertex_in, sFilename_output_in):

    if  os.path.exists(sFilename_flowline_in) and  os.path.exists(sFilename_vertex_in) : 
        pass
    else:
        print('The input file does not exist')
        return

    if os.path.exists(sFilename_output_in): 

        os.remove(sFilename_output_in)

    
    pDriver_geojson = ogr.GetDriverByName( "GeoJSON")

    pDataset_flowline = pDriver_geojson.Open(sFilename_flowline_in, 0)
    pDataset_vertex = pDriver_geojson.Open(sFilename_vertex_in, 0)   

    pLayer_flowline = pDataset_flowline.GetLayer(0)
    pSpatial_reference_a = pLayer_flowline.GetSpatialRef()
    nfeature_flowline = pLayer_flowline.GetFeatureCount()

    pLayer_vertex = pDataset_vertex.GetLayer(0)
    pSpatial_reference_b = pLayer_vertex.GetSpatialRef()
    nfeature_vertex = pLayer_vertex.GetFeatureCount()
    pLayerDefinition = pLayer_vertex.GetLayerDefn()

    comparison = pSpatial_reference_a.IsSame(pSpatial_reference_b)
    if(comparison != 1):
        iFlag_transform = 1
        transform = osr.CoordinateTransformation(pSpatial_reference_a, pSpatial_reference_b)
    else:
        iFlag_transform = 0

    pDataset_out = pDriver_geojson.CreateDataSource(sFilename_output_in)

    pLayerOut = pDataset_out.CreateLayer('flowline', pSpatial_reference_b, ogr.wkbMultiPoint)
    # Add one attribute
    pLayerOut.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    
    pLayerDefn = pLayerOut.GetLayerDefn()
    pFeatureOut = ogr.Feature(pLayerDefn)    
   
    lID_flowline = 0           

    aVertex_intersect=list()
    #for i in range (nfeature_mesh):
    for pFeature_flowline in pLayer_flowline:
       
        #pFeature_mesh= pLayer_mesh.GetFeature(i)
        pGeometry_flowline = pFeature_flowline.GetGeometryRef()        
        dummy0 = loads( pGeometry_flowline.ExportToWkt() )
        aCoords_gcs = dummy0.coords
        aCoords_gcs= np.array(aCoords_gcs)       

        if (iFlag_transform ==1): #projections are different
            pGeometry_flowline.Transform(transform)

        if (pGeometry_flowline.IsValid()):
            pass
        else:
            print('Geometry issue')

        #convert geometry to edge
        pGeometrytype_flowline = pGeometry_flowline.GetGeometryName()
        if(pGeometrytype_flowline == 'LINESTRING'):               
                     
            aFlowline_intersect = list()
            iFlag_intersected = 0 
            for j in range (nfeature_vertex):
            #for pFeature_flowline in pLayer_flowline:
                pFeature_vertex = pLayer_vertex.GetFeature(j)
                pGeometry_vertex = pFeature_vertex.GetGeometryRef()
                if (pGeometry_vertex.IsValid()):
                    pass
                else:
                    print('Geometry issue')

                iFlag_intersect = pGeometry_vertex.Intersects( pGeometry_flowline )
                if( iFlag_intersect == True):
                    iFlag_intersected = 1
                    pGeometry_intersect = pGeometry_vertex.Intersection(pGeometry_flowline)      
                    #add more process here to 
                    pGeometrytype_intersect = pGeometry_intersect.GetGeometryName()
                    if pGeometrytype_intersect == 'LINESTRING':
                        npoint = pGeometry_intersect.GetGeometryCount()
                        for i  in range(npoint): 
                            point = pGeometry_intersect.GetGeometryRef(i) 

                    else:
                        pass
                             
                    
                else:
                    pass
           
            
        else:
            pass

    pDataset_out = pLayerOut = None    

    return   aVertex_intersect




from pyflowline.algorithms.auxiliary.find_index_in_list import find_edge_in_list
def remove_duplicate_edge(aEdge_in):
    aEdge_out = list()
    nEdge = len(aEdge_in)
    for i in range(nEdge):
        pEdge = aEdge_in[i]
        iFlag_exist, lIndex = find_edge_in_list( aEdge_out,  pEdge)
        if iFlag_exist ==1:
            pass
        else:
            aEdge_out.append(pEdge)

        pass
    return aEdge_out


from abc import ABCMeta, abstractmethod

import copy

import json
from json import JSONEncoder
import numpy as np
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge


class FlowlineClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.float32):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson()) #lVertexID
        if isinstance(obj, pyedge):
            return obj.lEdgeID         
            
        return JSONEncoder.default(self, obj)

class pyflowline(object):
    __metaclass__ = ABCMeta 

    lFlowlineID=-1
    lIndex=-1
    lIndex_upstream=-1
    lIndex_downstream=-1

    iFlag_dam = 0
    lNHDPlusID=-1

    pVertex_start=None
    pVertex_end=None
    aEdge=None
    aVertex=None

    dLength=0.0
    dSinuosity=0.0

    iStream_segment=-1
    iStream_order =-1

    nEdge=0
    nVertex=0
    iFlag_right = 0
    iFlag_left = 0
    aFlowlineID_start_start = None
    aFlowlineID_start_end = None
    aFlowlineID_end_start = None
    aFlowlineID_end_end = None
    
    def __init__(self, aEdge):    
        self.aEdge = aEdge
        nEdge  = len(aEdge)
        self.nEdge = nEdge
        self.pVertex_start = aEdge[0].pVertex_start
        self.pVertex_end =  aEdge[ nEdge-1  ].pVertex_end
        nVertex = nEdge +1
        self.aVertex=list()
        for i in range(nEdge):
            self.aVertex.append( aEdge[i].pVertex_start )
            pass

        self.aVertex.append( aEdge[nEdge-1].pVertex_end )
        self.nVertex = nVertex
        self.dLength= self.calculate_length()
        self.aFlowlineID_start_start = list()
        self.aFlowlineID_start_end = list()
        self.aFlowlineID_end_start = list()
        self.aFlowlineID_end_end = list()
     
        return

    def calculate_length(self):
        dLength =0.0
        #loop though
        for edge in self.aEdge:
            edge.calculate_length()
            dLength = dLength + edge.dLength

        #assing
        self.dLength= dLength

        return dLength

    
    
    def check_upstream(self, other):
        iFlag_upstream =-1
        v0 = self.pVertex_start
        v1 = self.pVertex_end

        v2 = other.pVertex_start
        v3 = other.pVertex_end

        if v0 == v3:
            iFlag_upstream =1
        else:
            iFlag_upstream=0

        return iFlag_upstream

    def check_downstream(self, other):
        iFlag_downstream =-1
        v0 = self.pVertex_start
        v1 = self.pVertex_end
        v2 = other.pVertex_start
        v3 = other.pVertex_end
        if v1 == v2:
            iFlag_downstream =1
        else:
            iFlag_downstream=0

        return iFlag_downstream
    
    def reverse(self):
        '''
        reverse the direction of a flowline
        '''
        aVertex = self.aVertex 
        nVertex = self.nVertex
        aVertex_new = list()
        for i in range(nVertex-1,-1,-1) :
            aVertex_new.append( aVertex[i] )

        self.aVertex = aVertex_new
        nVertex  = len(aVertex)
        aEdge = list()
        for i in range(nVertex-1):
            pEdge = pyedge( self.aVertex[i], self.aVertex[i+1] )
            aEdge.append(pEdge)
            pass
        
        self.aEdge = aEdge
        nEdge = len(aEdge)
        self.pVertex_start = aEdge[0].pVertex_start
        self.pVertex_end =  aEdge[ nEdge-1  ].pVertex_end

    def merge_upstream(self, other):
        pFlowline_out = copy.deepcopy(other)    
        
        iFlag_dam1 = other.iFlag_dam
        pVertex_start1 = other.pVertex_start
        pVertex_end1 = other.pVertex_end
        nVertex1 = other.nVertex
        nEdge1 = other.nEdge

        pVertex_start2 = self.pVertex_start
        pVertex_end2 = self.pVertex_end
        nVertex2 = self.nVertex
        nEdge2 = self.nEdge
        iFlag_dam2 = self.iFlag_dam


        if pVertex_end1 == pVertex_start2:
            #this is the supposed operation because they should connect

            nVertex = nVertex1 + nVertex2 - 1
            nEdge = nVertex -1 
            aEdge = copy.deepcopy(other.aEdge )
            for i in range(nEdge2):
                aEdge.append( self.aEdge[i] )
                pass

            aVertex = copy.deepcopy(other.aVertex)
            for i in range(1, nVertex2):
                aVertex.append( self.aVertex[i] )
                pass
            
            pFlowline_out.iFlag_dam = max(iFlag_dam1, iFlag_dam2)
            pFlowline_out.aEdge = aEdge
            pFlowline_out.aVertex = aVertex
            pFlowline_out.nEdge = nEdge
            pFlowline_out.nVertex = nVertex
            pFlowline_out.dLength = self.dLength + other.dLength
            pFlowline_out.pVertex_start = pVertex_start1
            pFlowline_out.pVertex_end = pVertex_end2
            pass
        else:
            pass

        return pFlowline_out

    def calculate_flowline_sinuosity(self):
        pVertex_start = self.pVertex_start
        pVertex_end = self.pVertex_end
        dDistance = pVertex_start.calculate_distance(pVertex_end)
        self.dSinuosity = self.dLength / dDistance
        return

    def __eq__(self, other):                       
        iFlag_overlap = 0 
        nEdge1 = self.nEdge
        nEdge2 = other.nEdge
        if nEdge1 == nEdge2:
            for i in np.arange( nEdge1):
                pEdge1 = self.aEdge[i]
                pEdge2 = other.aEdge[i]
                if pEdge1 == pEdge2:
                    iFlag_overlap =1 
                else:
                    iFlag_overlap =0 
                    break                
            
        else:
            iFlag_overlap = 0

        return iFlag_overlap

    def __ne__(self, other):
        return not self.__eq__(other)
    
    def tojson(self):
        aSkip = ['aEdge', \
                'aVertex','aFlowlineID_start_start','aFlowlineID_start_end',
                'aFlowlineID_end_start','aFlowlineID_end_end']

        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        sJson = json.dumps(obj,  \
            sort_keys=True, \
                indent = 4, \
                    ensure_ascii=True, \
                        cls=FlowlineClassEncoder)
        return sJson
        


import os
import numpy as np
from osgeo import ogr, osr, gdal

from shapely.wkt import loads

from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_flowline

def read_flowline_shapefile(sFilename_shapefile_in):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """
    iReturn_code = 1
    if os.path.isfile(sFilename_shapefile_in):
        pass
    else:
        print('This shapefile does not exist: ', sFilename_shapefile_in )
        iReturn_code = 0
        return iReturn_code

    aFlowline=list()
    pDriver_shapefile = ogr.GetDriverByName('ESRI Shapefile')   
    pDataset_shapefile = pDriver_shapefile.Open(sFilename_shapefile_in, gdal.GA_ReadOnly)
    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
    pSpatialRef_shapefile = pLayer_shapefile.GetSpatialRef()
    pSpatial_reference_gcs = osr.SpatialReference()
    pSpatial_reference_gcs.ImportFromEPSG(4326)
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    comparison = pSpatialRef_shapefile.IsSame(pSpatial_reference_gcs)
    if(comparison != 1):
        iFlag_transform =1
        pTransform = osr.CoordinateTransformation(pSpatialRef_shapefile, pSpatial_reference_gcs)
    else:
        iFlag_transform =0
   
    lID = 0
    for pFeature_shapefile in pLayer_shapefile:        
        pGeometry_in = pFeature_shapefile.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        lNHDPlusID = int(pFeature_shapefile.GetField("NHDPlusID"))
        if (iFlag_transform ==1): #projections are different
            pGeometry_in.Transform(pTransform)
        if (pGeometry_in.IsValid()):
            pass
        else:
            print('Geometry issue')

        if(sGeometry_type == 'MULTILINESTRING'):
            aLine = ogr.ForceToLineString(pGeometry_in)
            for Line in aLine: 
                dummy = loads( Line.ExportToWkt() )
                aCoords = dummy.coords                
                dummy1= np.array(aCoords)
                pLine = convert_gcs_coordinates_to_flowline(dummy1)
                pLine.lIndex = lID
                pLine.lNHDPlusID= lNHDPlusID
                aFlowline.append(pLine)
                lID = lID + 1
               
        else:
            if sGeometry_type =='LINESTRING':
                dummy = loads( pGeometry_in.ExportToWkt() )
                aCoords = dummy.coords                
                dummy1= np.array(aCoords)
                pLine = convert_gcs_coordinates_to_flowline(dummy1)
                pLine.lIndex = lID
                pLine.lNHDPlusID= lNHDPlusID
                aFlowline.append(pLine)
                lID = lID + 1
                
            else:
                print(sGeometry_type)
                pass
        
        
    
    #we also need to spatial reference

    return aFlowline, pSpatialRef_shapefile

def read_flowline_shapefile_swat(sFilename_shapefile_in):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """
    aFlowline=list()
    pDriver_shapefile = ogr.GetDriverByName('ESRI Shapefile')   
    pDataset_shapefile = pDriver_shapefile.Open(sFilename_shapefile_in, gdal.GA_ReadOnly)
    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
    pSpatialRef_shapefile = pLayer_shapefile.GetSpatialRef()
    pSpatial_reference_gcs = osr.SpatialReference()
    pSpatial_reference_gcs.ImportFromEPSG(4326)
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    comparison = pSpatialRef_shapefile.IsSame(pSpatial_reference_gcs)
    if(comparison != 1):
        iFlag_transform =1
        pTransform = osr.CoordinateTransformation(pSpatialRef_shapefile, pSpatial_reference_gcs)
    else:
        iFlag_transform =0

    lID = 0
    for pFeature_shapefile in pLayer_shapefile:        
        pGeometry_in = pFeature_shapefile.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        if (iFlag_transform ==1): #projections are different
            pGeometry_in.Transform(pTransform)

        if (pGeometry_in.IsValid()):
            pass
        else:
            print('Geometry issue')

        if(sGeometry_type == 'MULTILINESTRING'):
            aLine = ogr.ForceToLineString(pGeometry_in)
            for Line in aLine: 
                dummy = loads( Line.ExportToWkt() )
                aCoords = dummy.coords      
                dummy1= np.array(aCoords)
                pLine = convert_gcs_coordinates_to_flowline(dummy1)
                pLine.lIndex = lID            
                aFlowline.append(pLine)
                lID = lID + 1
               
        else:
            if sGeometry_type =='LINESTRING':
                dummy = loads( pGeometry_in.ExportToWkt() )
                aCoords = dummy.coords               
                dummy1= np.array(aCoords)
                pLine = convert_gcs_coordinates_to_flowline(dummy1)
                pLine.lIndex = lID       
                aFlowline.append(pLine)
                lID = lID + 1
                
            else:
                print(sGeometry_type)
                pass
    

    return aFlowline, pSpatialRef_shapefile


def read_flowline_geojson(sFilename_geojson_in):
    """
    read a geojson flowline
    This function should be used for stream flowline only.
    """

    aFlowline=list()
    pDriver_geojson = ogr.GetDriverByName('GeoJSON')   
    pDataset_geojson = pDriver_geojson.Open(sFilename_geojson_in, gdal.GA_ReadOnly)
    pLayer_geojson = pDataset_geojson.GetLayer(0)
    pSpatialRef_geojson = pLayer_geojson.GetSpatialRef()
    ldefn = pLayer_geojson.GetLayerDefn()
    schema =list()
    for n in range(ldefn.GetFieldCount()):
        fdefn = ldefn.GetFieldDefn(n)
        schema.append(fdefn.name)
    if 'iseg' in schema:
        iFlag_segment = 1
    else:
        iFlag_segment = 0    
    if 'id' in schema:
        iFlag_id = 1
    else:
        iFlag_id = 0

    lID = 0
    for pFeature_geojson in pLayer_geojson:
        pGeometry_geojson = pFeature_geojson.GetGeometryRef()
        pGeometry_in = pFeature_geojson.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()
        if iFlag_segment ==1:
            iStream_segment = pFeature_geojson.GetField("iseg")
        else:
            iStream_segment = -1
        
        if iFlag_id ==1:
            lFlowlineID = pFeature_geojson.GetField("id")
        else:
            lFlowlineID = -1
        
        if sGeometry_type =='LINESTRING':
            dummy = loads( pGeometry_in.ExportToWkt() )
            aCoords = dummy.coords
            dummy1= np.array(aCoords)
            pLine = convert_gcs_coordinates_to_flowline(dummy1)
            pLine.lIndex = lID
            pLine.iStream_segment = iStream_segment
            pLine.lFlowlineID = lFlowlineID
            aFlowline.append(pLine)
            lID = lID + 1            
        else:
            print(sGeometry_type)
            pass        


    return aFlowline, pSpatialRef_geojson



import json
import numpy as np
from json import JSONEncoder

from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge
from pyflowline.classes.cell import pycell
from pyflowline.classes.flowline import pyflowline


class LinkClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, pyedge):
            return obj.lEdgeID
        if isinstance(obj, pyvertex):
            return obj.lVertexID
        if isinstance(obj, pyflowline):
            return obj.lFlowlineID
        if isinstance(obj, pycell):
            return obj.lCellID    
        return JSONEncoder.default(self, obj)

class pycelllink(object):
    lIndex=0
    pCell_start=None
    pCell_end=None
    pEdge_link = None
    def __init__(self, pCell_start_in, pCell_end_in, pEdge_link_in):
        self.pCell_start = pCell_start_in
        self.pCell_end = pCell_end_in
        self.pEdge_link = pEdge_link_in
        return
    
    def tojson(self):        
        sJson = json.dumps(self.__dict__, \
            ensure_ascii=True, \
                indent=4, \
                    cls=LinkClassEncoder)
        return sJson

#class pyhexagonlink(object):
#    
#    lIndex=0    
#    dLength=0.0
#    pHexagon_start=None
#    pHexagon_end=None
#    pEdge_link=None
#
#    def __init__(self, pHexagon_start_in, pHexagon_end_in, pEdge_link_in)#:   
#        self.pHexagon_start = pHexagon_start_in
#        self.pHexagon_end = pHexagon_end_in
#        self.pEdge_link = pEdge_link_in
#        return
#
#class pympaslink(object):
#    lIndex=0
#    
#    dLength=0.0
#    pMpas_start=None
#    pMpas_end=None
#    pEdge_link=None
#
#    def __init__(self, pMpas_start_in, pMpas_end_in, pEdge_link_in):   
#        self.pMpas_start = pMpas_start_in
#        self.pMpas_end = pMpas_end_in
#        self.pEdge_link = pEdge_link_in
#        return
#
#    def tojson(self):
#        
#        sJson = json.dumps(self.__dict__, \
#            ensure_ascii=True, \
#                indent=4, \
#                    cls=LinkClassEncoder)
#        return sJson
   
    




import os
import numpy as np
import json

def export_mesh_info_to_json(aCell_in, aFlowline_in, aCellID_outlet_iin, sFilename_json_in):
    """_summary_

    Args:
        aCell_in (_type_): _description_
        aFlowline_in (_type_): _description_
        aCellID_outlet_iin (_type_): _description_
        sFilename_json_in (_type_): _description_
    """
    if os.path.exists(sFilename_json_in): 
        os.remove(sFilename_json_in)
        pass
    ncell=len(aCell_in)
    nFlowline = len(aFlowline_in)
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        nEdge = pFlowline.nEdge
        nVertex = pFlowline.nVertex
        aEdge = pFlowline.aEdge
        iStream_segment = pFlowline.iStream_segment
        iStream_order = pFlowline.iStream_order
        for j in range(nEdge):
            pEdge = aEdge[j]
            pVertex_start = pEdge.pVertex_start
            pVertex_end = pEdge.pVertex_end
            for k in range(ncell):
                pVertex_center = aCell_in[k].pVertex_center
                if pVertex_center == pVertex_start:
                    aCell_in[k].iStream_segment_burned = iStream_segment
                    aCell_in[k].iStream_order_burned = iStream_order
                    for l in range(ncell):
                        pVertex_center2 = aCell_in[l].pVertex_center
                        lCellID = aCell_in[l].lCellID
                        if pVertex_center2 == pVertex_end:
                            aCell_in[k].lCellID_downstream_burned = lCellID
                            if lCellID in aCellID_outlet_iin:
                                aCell_in[l].iStream_segment_burned = iStream_segment
                                aCell_in[l].iStream_order_burned = iStream_order
                            break
                        
                        
        pass
    

    with open(sFilename_json_in, 'w', encoding='utf-8') as f:
        sJson = json.dumps([json.loads(ob.tojson()) for ob in aCell_in], indent = 4)        
        f.write(sJson)    
        f.close()
    
    return




from pyflowline.algorithms.auxiliary.find_index_in_list import add_unique_vertex

def find_flowline_vertex(aFlowline_in): 
    nFlowline = len(aFlowline_in) 
    aVertex=list()
    for i in range(0, nFlowline):      
        pFlowline = aFlowline_in[i]
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        aVertex, dummy = add_unique_vertex(aVertex, pVertex_start)
        aVertex, dummy = add_unique_vertex(aVertex, pVertex_end)
        pass

    return aVertex




import sys, os, stat
import numpy as np
from pathlib import Path
from shutil import copy2
import subprocess
import datetime


pDate = datetime.datetime.today()
sDate_default = "{:04d}".format(pDate.year) + "{:02d}".format(pDate.month) + "{:02d}".format(pDate.day)


def create_dggrid_case( iCase_index, \
    iResolution_index ,\
        sGrid_type,\
            sDate_in = None, \
        sFilename_crop_shapefile_in =None ):

    sCase = "{:03d}".format( iCase_index )
    sResolution = "{:02d}".format( iResolution_index )

    if sDate_in is not None:
        sDate = sDate_in 
    else:
        sDate = sDate_default

    if sFilename_crop_shapefile_in is not None:
        if os.path.isfile(sFilename_crop_shapefile_in):
            iFlag_crop =1
            sFilename_crop_shapefile = sFilename_crop_shapefile_in
        else:
            iFlag_crop = 0
    else:
        iFlag_crop = 0


    sCase_folder = sWorkspace_job + slash + sGrid_type + sResolution + slash + 'case'  +  sDate + sCase
    sFolder_out = sWorkspace_out  + slash + sGrid_type + sResolution + slash + 'case'  +  sDate + sCase
    sFilename_out = sFolder_out + slash + sGrid_type + sResolution 

    if (os.path.exists(sFolder_out)):
        sCommand = 'rm -rf '  + sFolder_out
        print(sCommand)
        p = subprocess.Popen(sCommand, shell= True)
        p.wait()
    Path(sFolder_out).mkdir(parents=True, exist_ok=True)

    if (os.path.exists(sCase_folder)):
        sCommand = 'rm -rf '  + sCase_folder
        print(sCommand)
        p = subprocess.Popen(sCommand, shell= True)
        p.wait()

    Path(sCase_folder).mkdir(parents=True, exist_ok=True)
    os.chdir(sCase_folder)
     #write configuration
    sFilename_config= sCase_folder + slash + sRegion +'.ini'
    ofs = open(sFilename_config, 'w')
    sLine = 'dggrid_operation GENERATE_GRID' + '\n'
    ofs.write(sLine)
    sLine = 'dggs_type ' + sGrid_type.upper() + '\n'
    ofs.write(sLine)
    sLine = 'dggs_res_spec ' + sResolution + '\n'
    ofs.write(sLine)

    if iFlag_crop ==1:
        sLine = 'clip_region_files ' + sFilename_crop_shapefile + '\n'
        ofs.write(sLine)
        sLine = 'clip_subset_type SHAPEFILE'  + '\n'
        ofs.write(sLine)

    else:
        pass

    sLine = 'update_frequency 10000000'+ '\n'
    ofs.write(sLine)
    sLine = 'cell_output_type SHAPEFILE' + '\n'
    ofs.write(sLine)
    sLine = 'cell_output_file_name ' + sFilename_out + '\n'
    ofs.write(sLine)
    sLine = 'densification 3' + '\n'
    ofs.write(sLine)
    sLine = 'max_cells_per_output_file 0'  + '\n'
    ofs.write(sLine)
    
    ofs.close()


    #writen normal run script
    sFilename_bash = sCase_folder + slash + 'run.sh'
    ofs = open(sFilename_bash, 'w')
    sLine = '#!/bin/bash\n'
    ofs.write(sLine)
    sLine = 'module load gcc/5.2.0\n'
    ofs.write(sLine)
    
    sLine = './dggrid ' + sRegion + '.ini' + '\n'
    ofs.write(sLine)
    ofs.close()
    os.chmod(sFilename_bash, stat.S_IREAD | stat.S_IWRITE | stat.S_IXUSR)

    #write job file
    sFilename_job = sCase_folder + slash + 'submit.job'
    ofs = open(sFilename_job, 'w')
    sLine = '#!/bin/bash\n'
    ofs.write(sLine)
    sLine = '#SBATCH -A br21_liao313\n'
    ofs.write(sLine)
    sLine = '#SBATCH  --job-name=dggrid' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH -t '+ sWalltime +':00:00' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH  --nodes=1' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH  --ntasks-per-node=1' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH  --partition=slurm' + '\n'
    ofs.write(sLine)
    sLine = '#SBATCH -o stdout.out\n'
    ofs.write(sLine)
    sLine = '#SBATCH -e stderr.err\n'
    ofs.write(sLine)
    sLine = '#SBATCH --mail-type=ALL\n'
    #ofs.write(sLine)
    sLine = '#SBATCH --mail-user=chang.liao@pnnl.gov\n'
    ofs.write(sLine)
    sLine = 'cd $SLURM_SUBMIT_DIR\n'
    ofs.write(sLine)
    sLine = 'module purge\n'
    ofs.write(sLine)
    sLine = 'module load gcc/5.2.0\n'
    ofs.write(sLine)
    
    sLine = './dggrid ' + sRegion + '.ini' + '\n'
    ofs.write(sLine)
    ofs.close()
   
    #copy execulate
    sFilename_new = sCase_folder + slash + 'dggrid'
    copy2(sFilename_dggrid, sFilename_new)
    os.chmod(sFilename_new, stat.S_IREAD | stat.S_IWRITE | stat.S_IXUSR)
    sCommand =  'sbatch submit.job' + '\n'
    sCommand = sCommand.lstrip()
    os.chdir(sCase_folder)
    
    p = subprocess.Popen(sCommand, shell= True)
    p.wait()
    return

if __name__ == '__main__':


    sRegion = 'conus'    

    sWorkspace_job = '/qfs/people/liao313/jobs/' + 'dggrid' + slash + sRegion + slash + 'simulation'
    sWorkspace_out = '/pic/scratch/liao313/04model/dggrid/' + slash + sRegion + slash + 'simulation'
    sFilename_dggrid = '/qfs/people/liao313/bin/dggrid'

    sFilename_boundary = '/qfs/people/liao313/data/dggrid/conus/vector/conus_simple.shp'
    iCase_index = 1
    sGrid_type='isea3h'
    for i  in np.arange(1, 15):
        sWalltime =  "{:02d}".format( iCase_index )
        iResolution_index = i 
        create_dggrid_case( iCase_index, \
                    iResolution_index ,  \
                        sGrid_type,
                    
                       sFilename_crop_shapefile_in=  sFilename_boundary)
        

       



    pass



import numpy as np
from pyflowline.classes.edge import pyedge
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.hexagon import pyhexagon
from pyflowline.classes.square import pysquare
from pyflowline.classes.latlon import pylatlon
from pyflowline.classes.mpas import pympas
from pyflowline.classes.tin import pytin
    
def convert_gcs_attributes_to_cell(iMesh_type_in, dLongitude_center_in, dLatitude_center_in, \
        aCoordinates_gcs_in, \
        aVertexID_in, \
        aEdgeID_in, \
        aVertexIndexOnEdge_in):  
        
    npoint = len(aVertexID_in)     
    aVertex=list()        
    aEdge=list()
    
    if iMesh_type_in == 1: #hexagon
        pHexagon = pyhexagon(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
        return pHexagon
    else:
        if iMesh_type_in == 2: #sqaure
            pSquare = pysquare(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
            return pSquare
        else:
            if iMesh_type_in == 3: #latlon
                pLatlon = pylatlon(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
                return pLatlon
            else:
                if iMesh_type_in == 4: #mpas
                    for i in range(npoint):
                        lon = aCoordinates_gcs_in[i][0]
                        lat = aCoordinates_gcs_in[i][1]
                        pVertex = dict()        
                        pVertex['dLongitude_degree'] =lon
                        pVertex['dLatitude_degree'] =lat
                        pVertex = pyvertex(pVertex)
                        pVertex.lVertexID = int(aVertexID_in[i])
                        aVertex.append(pVertex)

                    for j in range(npoint):
                        aVertexID_dummy = aVertexIndexOnEdge_in[j,:]
                        pVertex = dict()
                        dummy_index = np.where(aVertexID_in == aVertexID_dummy[0])     
                        pVertex['dLongitude_degree'] =aCoordinates_gcs_in[dummy_index,0]
                        pVertex['dLatitude_degree'] =aCoordinates_gcs_in[dummy_index,1]
                        pVertex_start = pyvertex(pVertex)
                        pVertex_start.lVertexID = int( aVertexID_dummy[0] )
                        dummy_index = np.where(aVertexID_in == aVertexID_dummy[1])     
                        pVertex['dLongitude_degree'] =aCoordinates_gcs_in[dummy_index,0]
                        pVertex['dLatitude_degree'] =aCoordinates_gcs_in[dummy_index,1]
                        pVertex_end = pyvertex(pVertex)
                        pVertex_end.lVertexID = int(aVertexID_dummy[1])
                        pEdge = pyedge( pVertex_start, pVertex_end )
                        pEdge.lEdgeID = int(aEdgeID_in[j])
                        aEdge.append(pEdge)

                    pMpas = pympas(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
                    return pMpas
                else:
                    if iMesh_type_in ==5: #tin
                        pTin = pytin(dLongitude_center_in, dLatitude_center_in, aEdge, aVertex)
                        return pTin
                        pass
                    else:
                        print('What mesh type are you using?')
                        return None
                    
  
def convert_pcs_attributes_to_cell(iMesh_type_in, \
    aCoordinates_pcs_in, \
    aVertexID_in, \
    aEdgeID_in,\
    aVertexIndexOnEdge_in):  

    npoint = len(aVertexID_in)     
    aVertex=list()        
    aEdge=list()
    
    if iMesh_type_in ==1: #hexagon
        pHexagon = pyhexagon( aEdge, aVertex)
        return pHexagon
    else:
        if iMesh_type_in ==2: #sqaure
            pSquare = pysquare( aEdge, aVertex)
            return pSquare
        else:
            if iMesh_type_in ==3: #latlon
                pLatlon = pylatlon( aEdge, aVertex)
                return pLatlon
            else:
                if iMesh_type_in ==4: #mpas
                    for i in range(npoint):
                        lon = aCoordinates_pcs_in[i][0]
                        lat = aCoordinates_pcs_in[i][1]
                        pVertex = dict()        
                        pVertex['dLongitude_degree'] =lon
                        pVertex['dLatitude_degree'] =lat
                        pVertex = pyvertex(pVertex)
                        pVertex.lVertexID = int(aVertexID_in[i])
                        aVertex.append(pVertex)

                    for j in range(npoint):
                        aVertexID_dummy = aVertexIndexOnEdge_in[j,:]
                        pVertex = dict()
                        dummy_index = np.where(aVertexID_in == aVertexID_dummy[0])     
                        pVertex['dLongitude_degree'] =aCoordinates_pcs_in[dummy_index,0]
                        pVertex['dLatitude_degree'] =aCoordinates_pcs_in[dummy_index,1]
                        pVertex_start = pyvertex(pVertex)
                        pVertex_start.lVertexID = int( aVertexID_dummy[0] )
                        dummy_index = np.where(aVertexID_in == aVertexID_dummy[1])     
                        pVertex['dLongitude_degree'] =aCoordinates_pcs_in[dummy_index,0]
                        pVertex['dLatitude_degree'] =aCoordinates_pcs_in[dummy_index,1]
                        pVertex_end = pyvertex(pVertex)
                        pVertex_end.lVertexID = int(aVertexID_dummy[1])
                        pEdge = pyedge( pVertex_start, pVertex_end )
                        pEdge.lEdgeID = int(aEdgeID_in[j])
                        aEdge.append(pEdge)

                    pMpas = pympas( aEdge, aVertex)
                    return pMpas
                else:
                    if iMesh_type_in ==5: #tin
                        pTin = pytin( aEdge, aVertex)
                        return pTin
                        pass
                    else:
                        print('What mesh type are you using?')
                        return None






import numpy as np
from pyflowline.algorithms.auxiliary.find_index_in_list import find_vertex_in_list
from pyflowline.algorithms.auxiliary.find_index_in_list import add_unique_vertex
from pyflowline.algorithms.auxiliary.check_head_water import check_head_water

def find_flowline_confluence(aFlowline_in, pVertex_outlet_in):    
    nFlowline = len(aFlowline_in) 
    aVertex=list()
    aIndex_headwater=list()
    aIndex_confluence=list()
    aIndex_middle =list()
    for i in range(0, nFlowline):      
        pFlowline = aFlowline_in[i]
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        aVertex, dummy = add_unique_vertex(aVertex, pVertex_start)
        aVertex, dummy = add_unique_vertex(aVertex, pVertex_end)
        pass

    nVertex=len(aVertex)
    aConnectivity  = np.full(  nVertex , 0, dtype=int )
    iFlag_first=1
    for i in range(nVertex):        
        pVertex = aVertex[i]
        dDiatance = pVertex.calculate_distance( pVertex_outlet_in) 
        if iFlag_first ==1:
            dDiatance_min = dDiatance                
            lIndex_outlet = i            
            iFlag_first=0
        else:
            if  dDiatance < dDiatance_min:
                dDiatance_min = dDiatance
                #found it
                lIndex_outlet = i        
                pass    
            else:
                pass
        
    aConnectivity[lIndex_outlet] =0 
    for i in range(0, nFlowline):      
        pFlowline = aFlowline_in[i]
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        if check_head_water(aFlowline_in, pVertex_start)==1:
            iFlag_exist, lIndex =  find_vertex_in_list(aVertex, pVertex_start)            
            aConnectivity[lIndex] = 1
            iFlag_exist, lIndex =  find_vertex_in_list(aVertex, pVertex_end)            
            aConnectivity[lIndex] = aConnectivity[lIndex] + 1        
        else:
            #it is either outlet, middle point, or confluence
            iFlag_exist, lIndex =  find_vertex_in_list(aVertex, pVertex_start)
            if iFlag_exist ==1:
                aConnectivity[lIndex] = aConnectivity[lIndex] + 1
            
            iFlag_exist, lIndex =  find_vertex_in_list(aVertex, pVertex_end)
            if iFlag_exist ==1:
                aConnectivity[lIndex] = aConnectivity[lIndex] + 1

    #reset outlet
    aConnectivity[lIndex_outlet] = 0
    for i in range(0, nVertex): 
        if (aConnectivity[i] >=3):
            aIndex_confluence.append(i)
        else:
            if (aConnectivity[i] ==1):
                aIndex_headwater.append(i)
            else:
                if (aConnectivity[i] ==2):
                    aIndex_middle.append(i)
                else:
                    #this is outlet
                    pass


    return aVertex, lIndex_outlet, aIndex_headwater, aIndex_middle, aIndex_confluence, aConnectivity



import os
import numpy as np
from osgeo import ogr, osr, gdal
from shapely.wkt import loads
from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_flowline

def read_nhdplus_flowline_shapefile_attribute(sFilename_shapefile_in):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """

    iReturn_code = 1
    if os.path.isfile(sFilename_shapefile_in):
        pass
    else:
        print('This shapefile does not exist: ', sFilename_shapefile_in )
        iReturn_code = 0
        return iReturn_code

    aNHDPlusID=list()
    pDriver_shapefile = ogr.GetDriverByName('ESRI Shapefile')
   
    pDataset_shapefile = pDriver_shapefile.Open(sFilename_shapefile_in, gdal.GA_ReadOnly)
    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
    pSpatialRef_shapefile = pLayer_shapefile.GetSpatialRef()            
    for pFeature_shapefile in pLayer_shapefile:        
        pGeometry_in = pFeature_shapefile.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()        
        lNHDPlusID = int(pFeature_shapefile.GetField("NHDPlusID"))
        aNHDPlusID.append(lNHDPlusID)        
    
    #we also need to spatial reference
    #aFromNode, aToNode, 
    return aNHDPlusID

def extract_nhdplus_flowline_shapefile_by_attribute(sFilename_shapefile_in, aAttribute_in):
    """_summary_

    Args:
        sFilename_shapefile_in (_type_): _description_
        aAttribute_in (_type_): _description_

    Returns:
        _type_: _description_
    """

    iReturn_code = 1

    if os.path.isfile(sFilename_shapefile_in):
        pass
    else:
        print('This shapefile does not exist: ', sFilename_shapefile_in )
        iReturn_code = 0
        return iReturn_code

    aFlowline =list()

    pDriver_shapefile = ogr.GetDriverByName('ESRI Shapefile')
   
    pDataset_shapefile = pDriver_shapefile.Open(sFilename_shapefile_in, gdal.GA_ReadOnly)
    pLayer_shapefile = pDataset_shapefile.GetLayer(0)
    pSpatialRef_shapefile = pLayer_shapefile.GetSpatialRef()

    pSpatial_reference_gcs = osr.SpatialReference()
    pSpatial_reference_gcs.ImportFromEPSG(4326)
    pSpatial_reference_gcs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    comparison = pSpatialRef_shapefile.IsSame(pSpatial_reference_gcs)
    if(comparison != 1):
        iFlag_transform =1
        pTransform = osr.CoordinateTransformation(pSpatialRef_shapefile, pSpatial_reference_gcs)
    else:
        iFlag_transform =0   

    lID = 0
    for pFeature_shapefile in pLayer_shapefile:        
        pGeometry_in = pFeature_shapefile.GetGeometryRef()
        sGeometry_type = pGeometry_in.GetGeometryName()       
        lNHDPlusID = int(pFeature_shapefile.GetField("NHDPlusID"))            
        if (iFlag_transform ==1): 
            pGeometry_in.Transform(pTransform)
        if (pGeometry_in.IsValid()):
            pass
        else:
            print('Geometry issue')
        if lNHDPlusID in aAttribute_in:
            if(sGeometry_type == 'MULTILINESTRING'):
                aLine = ogr.ForceToLineString(pGeometry_in)
                for Line in aLine: 
                    dummy = loads( Line.ExportToWkt() )
                    aCoords = dummy.coords                    
                    dummy1= np.array(aCoords)
                    pLine = convert_gcs_coordinates_to_flowline(dummy1)
                    pLine.lIndex = lID                    
                    aFlowline.append(pLine)
                    lID = lID + 1
            else:
                if sGeometry_type =='LINESTRING':
                    dummy = loads( pGeometry_in.ExportToWkt() )
                    aCoords = dummy.coords                   
                    dummy1= np.array(aCoords)
                    pLine = convert_gcs_coordinates_to_flowline(dummy1)
                    pLine.lIndex = lID                    
                    aFlowline.append(pLine)
                    lID = lID + 1
                else:
                    print(sGeometry_type)
                    pass
        else:
            pass

    return aFlowline

def track_nhdplus_flowline(aNHDPlusID_filter_in, aFromFlowline_in, aToFlowline_in, lNHDPlusID_in):    
    aNHDPlusID_dam_nonheadwater = list()
    def tag_downstream(lNHDPlusID_from):
        if lNHDPlusID_from in aNHDPlusID_filter_in:
            dummy_index = aNHDPlusID_filter_in.index(lNHDPlusID_from)
            pass
        else:
            if lNHDPlusID_from in aFromFlowline_in:

                if lNHDPlusID_from in aNHDPlusID_dam_nonheadwater:
                    pass
                else:
                    aNHDPlusID_dam_nonheadwater.append(lNHDPlusID_from)
                    dummy_index = np.where(aFromFlowline_in == lNHDPlusID_from )
                    nDownstream = dummy_index[0].size
                    for i in range(nDownstream):
                        lNHDPlusID_to = aToFlowline_in[dummy_index[0][i]   ]                    
                        if lNHDPlusID_to==0:
                            pass
                        else:
                       
                            tag_downstream(lNHDPlusID_to)
        return 

    if lNHDPlusID_in in aNHDPlusID_filter_in:
        dummy_index = aNHDPlusID_filter_in.index(lNHDPlusID_in)
          
    else:
        tag_downstream(lNHDPlusID_in)          
        aNHDPlusID_dam_nonheadwater.pop(0)    


    return  aNHDPlusID_dam_nonheadwater



import numpy as np 
from pyflowline.algorithms.auxiliary.check_head_water import check_head_water
def define_stream_order(aFlowline_in):
    nFlowline = len(aFlowline_in)
    aFlowline_out = list()
    
    if nFlowline == 0 :
        print ('data incomplete')
    else:       
        aStream_order = np.full(nFlowline, 0, dtype=int)    
        for i in range(nFlowline):
            pFlowline = aFlowline_in[i]
            pVertex_start=pFlowline.pVertex_start
            if check_head_water(aFlowline_in, pVertex_start)==1:     
                aStream_order[i] = 1
                pass       
     
        while aStream_order[0] == 0:
            for  i in range(nFlowline):
                if aStream_order[i] !=0:
                    continue

                pFlowline = aFlowline_in[i]
                iFlag_upstream_done = 1
                pVertex_start=pFlowline.pVertex_start
                pVertex_end=pFlowline.pVertex_end
                aStrord=list()
                for  j in range(nFlowline):
                    pFlowline2 = aFlowline_in[j]
                    pVertex_start2=pFlowline2.pVertex_start
                    pVertex_end2=pFlowline2.pVertex_end
                    if pVertex_start == pVertex_end2:
                        if aStream_order[j] == 0:
                            iFlag_upstream_done=0
                            break
                        else:
                            aStrord.append( aStream_order[j]  )
                
                if iFlag_upstream_done==1:
                    #get unique value
                    dummy = np.array(aStrord)
                    dummy1 = np.unique(dummy)

                    if len(dummy1) == 1: #all upstreams have the same order
                        aStream_order[i] = aStrord[0] + 1
                    else:
                        aStream_order[i] = np.max(dummy)    

        for i in range(nFlowline):
            pFlowline = aFlowline_in[i]      
            pFlowline.iStream_order =    aStream_order[i]   
            aFlowline_out.append(pFlowline)

    return aFlowline_out, aStream_order


import os
import numpy as np
from osgeo import ogr, osr
from shapely.wkt import loads
from pyflowline.classes.vertex import pyvertex
from pyflowline.algorithms.auxiliary.find_index_in_list import find_vertex_in_list

def intersect_flowline_with_flowline( sFilename_flowline_a_in, sFilename_flowline_b_in, sFilename_output_in):
    if  os.path.exists(sFilename_flowline_a_in) and  os.path.exists(sFilename_flowline_b_in) : 
        pass
    else:
        print('The input file does not exist')
        return

    if os.path.exists(sFilename_output_in): 
        os.remove(sFilename_output_in)
    
    pDriver_geojson = ogr.GetDriverByName( "GeoJSON")
    pDataset_flowline_a = pDriver_geojson.Open(sFilename_flowline_a_in, 0)
    pDataset_flowline_b = pDriver_geojson.Open(sFilename_flowline_b_in, 0)   
    pLayer_flowline_a = pDataset_flowline_a.GetLayer(0)
    pSpatial_reference_a = pLayer_flowline_a.GetSpatialRef()
    nfeature_flowline_a = pLayer_flowline_a.GetFeatureCount()
    pLayer_flowline_b = pDataset_flowline_b.GetLayer(0)
    pSpatial_reference_b = pLayer_flowline_b.GetSpatialRef()
    nfeature_flowline_b = pLayer_flowline_b.GetFeatureCount()
    pLayerDefinition = pLayer_flowline_b.GetLayerDefn()
    schema =list()
    for n in range(pLayerDefinition.GetFieldCount()):
        fdefn = pLayerDefinition.GetFieldDefn(n)
        schema.append(fdefn.name)
    if 'id' in schema:
        iFlag_id = 1
    else:
        iFlag_id = 0   
    
    comparison = pSpatial_reference_a.IsSame(pSpatial_reference_b)
    if(comparison != 1):
        iFlag_transform = 1
        transform = osr.CoordinateTransformation(pSpatial_reference_a, pSpatial_reference_b)
    else:
        iFlag_transform = 0

    pDataset_out = pDriver_geojson.CreateDataSource(sFilename_output_in)

    pLayerOut = pDataset_out.CreateLayer('flowline', pSpatial_reference_b, ogr.wkbMultiPoint)
    # Add one attribute
    pLayerOut.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    
    pLayerDefn = pLayerOut.GetLayerDefn()
    pFeatureOut = ogr.Feature(pLayerDefn)    
   
    lVertexID = 0           

    aVertex_intersect=list()
    #for i in range (nfeature_mesh):
    for pFeature_flowline_a in pLayer_flowline_a:
       
        #pFeature_mesh= pLayer_mesh.GetFeature(i)
        pGeometry_flowline_a = pFeature_flowline_a.GetGeometryRef()        
        dummy0 = loads( pGeometry_flowline_a.ExportToWkt() )
        aCoords_gcs = dummy0.coords
        aCoords_gcs= np.array(aCoords_gcs)       

        if (iFlag_transform ==1): #projections are different
            pGeometry_flowline_a.Transform(transform)

        if (pGeometry_flowline_a.IsValid()):
            pass
        else:
            print('Geometry issue')

        #convert geometry to edge
        pGeometrytype_flowline_a = pGeometry_flowline_a.GetGeometryName()
        if(pGeometrytype_flowline_a == 'LINESTRING'):            
            
                     
            aFlowline_intersect = list()
            iFlag_intersected = 0 
            for j in range (nfeature_flowline_b):
            #for pFeature_flowline in pLayer_flowline:
                pFeature_flowline_b = pLayer_flowline_b.GetFeature(j)
                pGeometry_flowline_b = pFeature_flowline_b.GetGeometryRef()

                if iFlag_id ==1:
                    lFlowlineID = pFeature_flowline_b.GetField("id")
                else:
                    lFlowlineID = -1

                if (pGeometry_flowline_b.IsValid()):
                    pass
                else:
                    print('Geometry issue')

                iFlag_intersect = pGeometry_flowline_b.Intersects( pGeometry_flowline_a )
                if( iFlag_intersect == True):
                    iFlag_intersected = 1
                    pGeometry_intersect = pGeometry_flowline_b.Intersection(pGeometry_flowline_a)      
                    #add more process here to 
                    pGeometrytype_intersect = pGeometry_intersect.GetGeometryName()
                    
                    if pGeometrytype_intersect == 'MULTIPOINT':
                        npoint = pGeometry_intersect.GetGeometryCount()
                        for i  in range(npoint): 
                            point = pGeometry_intersect.GetGeometryRef(i)                            

                            point0= dict()   
                            point0['dLongitude_degree'] = point.GetX()
                            point0['dLatitude_degree'] = point.GetY()
                            pVertex=pyvertex(point0)
                            pVertex.lFlowlineID = lFlowlineID
                            iFlag_exist, lIndex = find_vertex_in_list( aVertex_intersect,  pVertex)
                            if iFlag_exist ==1:
                                pass
                            else:                       
                                aVertex_intersect.append(pVertex)
                                pFeatureOut.SetGeometry(point)
                                pFeatureOut.SetField("id", lVertexID)         
                                pLayerOut.CreateFeature(pFeatureOut)    
                                lVertexID = lVertexID + 1

                    else:
                        #this branch possible has error, report an issue if crash
                        point= dict()   
                        point['dLongitude_degree'] = pGeometry_intersect.GetX()
                        point['dLatitude_degree'] = pGeometry_intersect.GetY()
                        pVertex=pyvertex(point)
                        pVertex.lFlowlineID = lFlowlineID
                        iFlag_exist, lIndex = find_vertex_in_list( aVertex_intersect,  pVertex)
                        if iFlag_exist ==1:
                            pass
                        else:                
                            aVertex_intersect.append(pVertex)
                            pFeatureOut.SetGeometry(pGeometry_intersect)
                            pFeatureOut.SetField("id", lVertexID)         
                            pLayerOut.CreateFeature(pFeatureOut)    
                            lVertexID = lVertexID + 1                                        
                    
                else:
                    pass           
            
        else:
            pass

    pDataset_out = pLayerOut = None    

    return   aVertex_intersect


from abc import ABCMeta
import numpy as np
import json
from json import JSONEncoder
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.flowline import pyflowline
from pyflowline.algorithms.auxiliary.gdal_functions import calculate_distance_based_on_lon_lat
from pyflowline.algorithms.auxiliary.gdal_functions import  calculate_angle_betwen_vertex

class ConfluenceClassEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()        
        if isinstance(obj, np.float32):
            return float(obj)        
        if isinstance(obj, list):
            pass  
        if isinstance(obj, pyvertex):
            return json.loads(obj.tojson())       
        if isinstance(obj, pyflowline):
            return obj.lFlowlineID
        
        return JSONEncoder.default(self, obj)

class pyconfluence():
    __metaclass__ = ABCMeta  
    lIndex=-1 
    lConfluenceID=-1    
    pVertex_confluence=None    
    pFlowline_downstream=None
    aFlowline_upstream=list()
    dAngle_upstream=0.0    
    def __init__(self, pVertex_center, aFlowline_upstream_in, pFlowline_downstream_in):        
        try:     
            self.pVertex_confluence      = pVertex_center       
            self.aFlowline_upstream      = aFlowline_upstream_in  
            self.pFlowline_downstream      = pFlowline_downstream_in  
            
        except:
            print('Initialization of confluence failed!')
        
        return
    
    def calculate_branching_angle(self):        
        #normally there are 2 edges meet at confluence        
        if len(self.aFlowline_upstream)==2:
            pFlowline1 = self.aFlowline_upstream[0]
            pFlowline2 = self.aFlowline_upstream[1]
            nedge1 = pFlowline1.nEdge
            nedge2 = pFlowline2.nEdge                
            x1 = pFlowline1.aEdge[nedge1-1].pVertex_start.dLongitude_degree
            y1 = pFlowline1.aEdge[nedge1-1].pVertex_start.dLatitude_degree
            x2 = self.pVertex_confluence.dLongitude_degree
            y2 = self.pVertex_confluence.dLatitude_degree
            x3 = pFlowline2.aEdge[nedge2-1].pVertex_start.dLongitude_degree
            y3 = pFlowline2.aEdge[nedge2-1].pVertex_start.dLatitude_degree
            self.dAngle_upstream = calculate_angle_betwen_vertex(x1, y1, x2, y2, x3, y3)
        else:
            print('multiple upstream')
            print(len(self.aFlowline_upstream))
            self.dAngle_upstream=0.0

        return self.dAngle_upstream    
    
    def tojson(self):
        aSkip = ['aFlowline_upstream']
        obj = self.__dict__.copy()
        for sKey in aSkip:
            obj.pop(sKey, None)
        sJson = json.dumps(obj,  \
                sort_keys=True, \
                indent = 4, \
                ensure_ascii=True, \
                cls=ConfluenceClassEncoder)
        return sJson


import os 
import datetime
import json
from pyflowline.classes.pycase import flowlinecase

pDate = datetime.datetime.today()
sDate_default = "{:04d}".format(pDate.year) + "{:02d}".format(pDate.month) + "{:02d}".format(pDate.day)

def pyflowline_read_model_configuration_file(sFilename_configuration_in,   \
    iFlag_standalone_in= None, \
        iFlag_use_mesh_dem_in=None, \
        iCase_index_in=None,   \
            dResolution_degree_in = None,  \
            dResolution_meter_in = None,  \
                sMesh_type_in = None,  \
                sModel_in = None, \
                    sDate_in = None,\
                    sWorkspace_output_in = None):
    """read a model configuration

    Args:
        sFilename_configuration_in (str): _description_
        iFlag_standalone_in (int, optional): _description_. Defaults to None.
        iFlag_use_mesh_dem_in (int, optional): _description_. Defaults to None.
        iCase_index_in (int, optional): _description_. Defaults to None.
        dResolution_degree_in (float, optional): _description_. Defaults to None.
        dResolution_meter_in (float, optional): _description_. Defaults to None.
        sMesh_type_in (str, optional): _description_. Defaults to None.
        sModel_in (str, optional): _description_. Defaults to None.
        sDate_in (str, optional): _description_. Defaults to None.
        sWorkspace_output_in (str, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """

    if not os.path.isfile(sFilename_configuration_in):
        print(sFilename_configuration_in + ' does not exist')
        return
    
    # Opening JSON file
    with open(sFilename_configuration_in) as json_file:
        aConfig = json.load(json_file)   
    
    if iCase_index_in is not None:        
        iCase_index = iCase_index_in
    else:       
        iCase_index = int( aConfig['iCase_index'])
    
    if iFlag_standalone_in is not None:        
        iFlag_standalone = iFlag_standalone_in
    else:       
        iFlag_standalone = int( aConfig['iFlag_standalone'])

    if iFlag_use_mesh_dem_in is not None:        
        iFlag_use_mesh_dem = iFlag_use_mesh_dem_in
    else:       
        iFlag_use_mesh_dem = int(aConfig['iFlag_use_mesh_dem'])

    if sMesh_type_in is not None:
        sMesh_type = sMesh_type_in
    else:
        sMesh_type = aConfig['sModel']
        pass
        
    if sModel_in is not None:
        sModel = sModel_in
    else:
        sModel = aConfig['sModel']
        pass

    if sDate_in is not None:
        sDate = sDate_in
    else:
        sDate = aConfig['sDate']
        pass

    if dResolution_degree_in is not None:
        dResolution_degree = dResolution_degree_in
    else:
        dResolution_degree = aConfig['dResolution_degree']
        pass

    if dResolution_meter_in is not None:
        dResolution_meter = dResolution_meter_in
    else:
        dResolution_meter = aConfig['dResolution_meter']
        pass

    if sWorkspace_output_in is not None:
        sWorkspace_output = sWorkspace_output_in
    else:
        sWorkspace_output = aConfig['sWorkspace_output']
        pass

    
    aConfig['iCase_index'] = iCase_index
    aConfig['iFlag_standalone'] = iFlag_standalone
    aConfig['iFlag_use_mesh_dem'] = iFlag_use_mesh_dem
    aConfig['dResolution_degree'] = dResolution_degree
    aConfig['dResolution_meter'] = dResolution_meter

    aConfig['sDate'] = sDate
    aConfig['sModel'] = sModel
    aConfig['sWorkspace_output'] = sWorkspace_output
   
    

    #based on global variable, a few variables are calculate once
    #calculate the modflow simulation period
    #https://docs.python.org/3/library/datetime.html#datetime-objects
   
   
    
    #aConfig
    
    #simulation
    
    oPyflowline = flowlinecase(aConfig)
   
    
    return oPyflowline


from pyflowline.classes.edge import pyedge

def connect_disconnect_flowline(aFlowline_in, aVertex_in, aThreshold_in):
    """[summary]

    Args:
        aFlowline_in ([type]): [description]
        aVertex_in ([type]): [description]
        aThreshold_in ([type]): [description]

    Returns:
        [type]: [description]
    """
    aFlowline_out = list()
    nFlowline = len(aFlowline_in)
    nvertex = len(aVertex_in)
    for j in range(nFlowline):
        pFlowline = aFlowline_in[j]
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        iFlag_found = 0
        for i in range(nvertex):
            if iFlag_found ==1:
                break
                
            pVertex=aVertex_in[i]
            dDistance1 = pVertex.calculate_distance(pVertex_start)
            dDistance2 = pVertex.calculate_distance(pVertex_end)
            
            if dDistance1 < aThreshold_in[i]:                
                pFlowline.aVertex.insert(0, pVertex)
                pEdge = pyedge(pVertex, pFlowline.aEdge[0].pVertex_start)
                pFlowline.aEdge.insert(0, pEdge)
                #update count
                pFlowline.nVertex = pFlowline.nVertex + 1
                pFlowline.nEdge = pFlowline.nEdge + 1                
                pFlowline.pVertex_start = pVertex
                iFlag_found = 1
                break
                
            else:
                if dDistance2 < aThreshold_in[i]:                    
                    nvertex2 = pFlowline.nVertex
                    nedge2 = pFlowline.nEdge
                    pFlowline.aVertex.insert(nvertex2, pVertex)
                    pEdge = pyedge(pFlowline.aEdge[nedge2-1].pVertex_end, pVertex )
                    pFlowline.aEdge.insert(nedge2, pEdge)
                    #update count
                    pFlowline.nVertex = pFlowline.nVertex + 1
                    pFlowline.nEdge = pFlowline.nEdge + 1
                    pFlowline.pVertex_end = pVertex
                    iFlag_found = 1
                    break
                    
                else:
                    pass

            pass
    
        aFlowline_out.append(pFlowline)
        pass

    return aFlowline_out

           

            

  


import os
import json
from osgeo import ogr, osr
from shapely.geometry import Point, LineString
from pyflowline.classes.edge import pyedge
from pyflowline.classes.link import pycelllink

def export_flowline_to_json( aFlowline_in, \
    sFilename_json_in, \
    iFlag_projected_in= None, \
    pSpatial_reference_in=None, \
    aAttribute_field=None,\
    aAttribute_data=None,\
    aAttribute_dtype=None):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """

    if os.path.exists(sFilename_json_in): 
        os.remove(sFilename_json_in)
        pass

    nFlowline = len(aFlowline_in)
    if iFlag_projected_in is None:
        iFlag_projected_in = 0
    else:
        iFlag_projected_in = 1

    if  pSpatial_reference_in is None:        
        pSpatial_reference_in = osr.SpatialReference()  
        pSpatial_reference_in.ImportFromEPSG(4326)    # WGS84 lat/lon
    else:
        pass

    if aAttribute_field is not None and aAttribute_data is not None and aAttribute_dtype is not None:
        iFlag_attribute = 1

        nAttribute1 = len(aAttribute_field)
        nAttribute2 = len(aAttribute_data)
        nAttribute3 = len(aAttribute_dtype)
        nAttribute4 = len(aAttribute_data[0])
        if nAttribute3 != nAttribute1 or nAttribute1 != nAttribute2 or nFlowline!= nAttribute4:
            print('The attribute is not correct, please check!')
            return
        else:
            iFlag_attribute = 1
    else:
        iFlag_attribute=0
        pass


    pDriver_json = ogr.GetDriverByName('GeoJSON')
    
    pDataset_json = pDriver_json.CreateDataSource(sFilename_json_in)  
    

    pLayer_json = pDataset_json.CreateLayer('flowline', pSpatial_reference_in, ogr.wkbLineString)
    # Add one attribute
    pLayer_json.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution

    #add the other fields
    if iFlag_attribute ==1:
        for i in range(nAttribute1):
            sField = aAttribute_field[i]
            dtype = aAttribute_dtype[i]
            if dtype == 'int':
                pLayer_json.CreateField(ogr.FieldDefn(sField, ogr.OFTInteger64))
                pass
            else:
                pLayer_json.CreateField(ogr.FieldDefn(sField, ogr.OFTReal))
                pass
        
    
    pLayerDefn = pLayer_json.GetLayerDefn()
    pFeature_out = ogr.Feature(pLayerDefn)

    lID = 0
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        dummy =pFlowline.aVertex
        aPoint=list()
        for j in dummy:
            if iFlag_projected_in ==1:
                aPoint.append( Point( j.dx, j.dy ) )                
                pass
            else:
                aPoint.append( Point( j.dLongitude_degree, j.dLatitude_degree ) )
                pass

        dummy1= LineString( aPoint )
        pGeometry_out = ogr.CreateGeometryFromWkb(dummy1.wkb)
        pFeature_out.SetGeometry(pGeometry_out)
   
        pFeature_out.SetField("id", lID)
        if iFlag_attribute ==1:
            for k in range(nAttribute1):
                sField = aAttribute_field[k]
                dtype = aAttribute_dtype[k]
                dummy = aAttribute_data[k]
                if dtype == 'int':
                    pFeature_out.SetField(sField, int(dummy[i]))
                else:
                    pFeature_out.SetField(sField, float(dummy[i]))
        
        # Add new pFeature_shapefile to output Layer
        pLayer_json.CreateFeature(pFeature_out)        
        lID =  lID + 1
        pass
        
    pDataset_json.FlushCache()
    pDataset_json = pLayer_json = pFeature_out  = None    

    return

    

def export_flowline_to_shapefile(iFlag_projected_in, aFlowline_in, pSpatial_reference_in, \
    sFilename_shapefile_in, \
    aAttribute_field=None,\
    aAttribute_data=None,\
        aAttribute_dtype=None):
    """
    convert a shpefile to json format.
    This function should be used for stream flowline only.
    """

    if os.path.exists(sFilename_shapefile_in): 
        #delete it if it exists
        os.remove(sFilename_shapefile_in)
        pass

    nFlowline = len(aFlowline_in)

    if aAttribute_field is not None and aAttribute_data is not None and aAttribute_dtype is not None:
        iFlag_attribute = 1

        nAttribute1 = len(aAttribute_field)
        nAttribute2 = len(aAttribute_data)
        nAttribute3 = len(aAttribute_dtype)
        nAttribute4 = len(aAttribute_data[0])
        if nAttribute3 != nAttribute1 or nAttribute1 != nAttribute2 or nFlowline!= nAttribute4:
            print('The attribute is not correct, please check!')
            return
        else:
            iFlag_attribute = 1
    else:
        iFlag_attribute=0
        pass

    pDriver_shapefile = ogr.GetDriverByName('ESRI Shapefile')
    pDataset_shapefile = pDriver_shapefile.CreateDataSource(sFilename_shapefile_in)      
    pLayer_shapefile = pDataset_shapefile.CreateLayer('flowline', pSpatial_reference_in, ogr.wkbLineString)
    # Add one attribute
    pLayer_shapefile.CreateField(ogr.FieldDefn('id', ogr.OFTInteger64)) #long type for high resolution
    #add the other fields
    if iFlag_attribute ==1:
        for i in range(nAttribute1):
            sField = aAttribute_field[i]
            dtype = aAttribute_dtype[i]
            if dtype == 'int':
                pLayer_shapefile.CreateField(ogr.FieldDefn(sField, ogr.OFTInteger64))
                pass
            else:
                pLayer_shapefile.CreateField(ogr.FieldDefn(sField, ogr.OFTReal))
                pass
        
    
    pLayerDefn = pLayer_shapefile.GetLayerDefn()
    pFeature_out = ogr.Feature(pLayerDefn)

    lID = 0
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        dummy =pFlowline.aVertex
        aPoint=list()
        for j in dummy:
            if iFlag_projected_in ==1:
                aPoint.append( Point( j.dx, j.dy ) )                
                pass
            else:
                aPoint.append( Point( j.dLongitude_degree, j.dLatitude_degree ) )
                pass

        dummy1= LineString( aPoint )
        pGeometry_out = ogr.CreateGeometryFromWkb(dummy1.wkb)
        pFeature_out.SetGeometry(pGeometry_out)
   
        pFeature_out.SetField("id", lID)
        if iFlag_attribute ==1:
            for k in range(nAttribute1):
                sField = aAttribute_field[k]
                dtype = aAttribute_dtype[k]
                dummy = aAttribute_data[k]
                if dtype == 'int':
                    pFeature_out.SetField(sField, int(dummy[i]))
                else:
                    pFeature_out.SetField(sField, float(dummy[i]))
        
        # Add new pFeature_shapefile to output Layer
        pLayer_shapefile.CreateFeature(pFeature_out)        
        lID =  lID + 1
        pass
        
    pDataset_shapefile.FlushCache()
    pDataset_shapefile = pLayer_shapefile = pFeature_out  = None    

    return

  

def export_flowline_info_to_json(aCell, aCell_intersect_in, aFlowline_in, sFilename_json_out):    
    #export the flowline topology to json
    ncell= len(aCell_intersect_in)
    nflowline= len(aFlowline_in)
    aLink =list()   
    for i in range(1, nflowline+1):
        pFlowline = aFlowline_in[i-1]
        nVertex = pFlowline.nVertex
        nEdge = pFlowline.nEdge
        for j in range(1, nEdge+1):         
            pEdge = pFlowline.aEdge[j-1]
            pVertex_start = pEdge.pVertex_start
            pVertex_end = pEdge.pVertex_end
            for k in range(ncell):
                if aCell_intersect_in[k].pVertex_center == pVertex_start:
                    pMpas_start = aCell_intersect_in[k]
                    pass
                if aCell_intersect_in[k].pVertex_center == pVertex_end:
                    pMpas_end = aCell_intersect_in[k]
                    pass
            pEdge_link = pyedge(pVertex_start, pVertex_end)  
            pLink = pycelllink(pMpas_start, pMpas_end, pEdge_link)
            aLink.append(pLink)

    with open(sFilename_json_out, 'w', encoding='utf-8') as f:
        sJson = json.dumps([json.loads(ob.tojson()) for ob in aLink], indent = 4)
        f.write(sJson)    
        f.close()
                
    
    return

    
   


from pyflowline.classes.flowline import pyflowline
def split_flowline_to_edge(aFlowline_in):
    aFlowline_out = list()
    aEdge_out=list()
    nFlowline = len(aFlowline_in)
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        iStream_order = pFlowline.iStream_order
        nVertex = pFlowline.nVertex
        nEdge = pFlowline.nEdge
        iPart = 0        
        for j in range(nEdge):
            pEdge = pFlowline.aEdge[j]
            aEdge = list()
            aEdge.append(pEdge)
            aEdge_out.append(pEdge)            
            pFlowline1 = pyflowline(aEdge)      
            aFlowline_out.append(pFlowline1)    
            pass
        pass   
    return aFlowline_out, aEdge_out




import copy
import numpy as np
from pyflowline.formats.convert_coordinates import convert_gcs_coordinates_to_flowline
from pyflowline.algorithms.auxiliary.find_index_in_list import check_if_duplicates

def remove_returning_flowline(iMesh_type_in, aCell_intersect_in, pVertex_outlet_in):    
    aFlowline_out=list()    
    nCell =  len(aCell_intersect_in)
    def simplify_list(aCell_flowline_in):
        aCell_flowline_out = copy.deepcopy(aCell_flowline_in)
        nCell2 = len(aCell_flowline_out)          
        iFlag_unique = check_if_duplicates(aCell_flowline_out)
        if iFlag_unique == 1:
            return aCell_flowline_out
        else:
            for i in range(nCell2):
                elem=aCell_flowline_out[i]
                if aCell_flowline_out.count(elem) > 1:
                    dummy = [i for i, x in enumerate(aCell_flowline_out) if x == elem]                    
                    start = dummy[0]
                    end = dummy[-1]
                    del aCell_flowline_out[start: end]                     
                    break                    
                else:
                    pass

                pass             
            
            aCell_flowline_out = simplify_list(aCell_flowline_out)
        return   aCell_flowline_out

    def retrieve_flowline_intersect_index(iSegment_in, iStream_order_in, pVertex_end_in):
        iFlag_found = 1
        pVertex_end_current = pVertex_end_in
        aCell_flowline=list()  
        aSegment_upstream = list()
        aVertex_end_upstream = list()
        aStream_order = list()
        iFlag_skip = 0
        iFlag_previous_overlap=0
        while iFlag_found == 1:
            iFlag_found = 0             
            for j in range(nCell):
                pCell = aCell_intersect_in[j]
                lCellID = pCell.lCellID
                aFlowline= pCell.aFlowline
                nFlowline = len(aFlowline)
                for i in range(nFlowline):
                    pFlowline = aFlowline[i]
                    pVertex_start = pFlowline.pVertex_start
                    pVertex_end = pFlowline.pVertex_end
                    iStream_segment = pFlowline.iStream_segment
                    iStream_order = pFlowline.iStream_order
                    if pVertex_end == pVertex_end_current:
                        if iStream_segment == iSegment_in:                            
                            dLength = pFlowline.dLength
                            iFlag_found2, dummy2 = pCell.which_edge_cross_this_vertex(pVertex_start)
                            iFlag_found3, dummy3 = pCell.which_edge_cross_this_vertex(pVertex_end)

                            if iFlag_found2 == 1 and iFlag_found3 == 1: #on the edge
                                #same edge
                                if dummy2.is_overlap(dummy3) ==1:                                    
                                    pVertex_end_current = pVertex_start    
                                    aCell_flowline.append(lCellID)
                                    iFlag_found = 1                
                                    iFlag_previous_overlap =1
                                    
                                    pass
                                else:
                                    if dLength < (0.1*pCell.dLength) : #because it taks a short cut                                        
                                        pVertex_end_current = pVertex_start   
                                        iFlag_found = 1                                          
                                        if iFlag_previous_overlap ==1: 
                                            aCell_flowline.append(lCellID)              
                                            iFlag_previous_overlap=0
                                            pass
                                        else:
                                            iFlag_previous_overlap=1
                                            pass                                             
                                        pass
                                    else:
                                        pVertex_end_current = pVertex_start    
                                        aCell_flowline.append(lCellID)
                                        iFlag_found = 1                                      
                                        pass

                                    pass                                

                                pass

                            else:
                                pVertex_end_current = pVertex_start    
                                aCell_flowline.append(lCellID)                                
                                iFlag_found = 1
                                iFlag_previous_overlap=0
                                pass                            
                                
                            break
                            pass
                        else:
                            iFlag_found2, dummy2 = pCell.which_edge_cross_this_vertex(pVertex_start)
                            iFlag_found3, dummy3 = pCell.which_edge_cross_this_vertex(pVertex_end)
                            if pVertex_end == pVertex_end_in: 
                                iFlag_found = 1
                                pass
                            else:
                                #this is a new upstream segment location
                                iFlag_found = 0
                                #save the upstream information for new step
                                aSegment_upstream.append(iStream_segment)
                                aVertex_end_upstream.append(pVertex_end)
                                aStream_order.append(iStream_order)                               
                            
                            pass
                    pass

                if iFlag_found == 1:
                    break
        
        #should have finished search
        if iFlag_found == 0:
            #reverse 
            aCell_flowline = aCell_flowline[::-1]
            #simplify list         
            aCell_simple = simplify_list(aCell_flowline)
            #save the output
            nCell3 = len(aCell_simple)
            aCoordinates = list()            
            if nCell3 >1:
                for i in range(nCell3):
                    lCellID = aCell_simple[i]
                    for j in range( nCell):
                        if aCell_intersect_in[j].lCellID == lCellID:    
                            x = aCell_intersect_in[j].dLongitude_center_degree
                            y = aCell_intersect_in[j].dLatitude_center_degree                
                            aCoordinates.append([x,y])
                    pass
                pFlowline = convert_gcs_coordinates_to_flowline(aCoordinates)
                pFlowline.iStream_segment = iSegment_in
                pFlowline.iStream_order=iStream_order_in
                aFlowline_out.append(pFlowline)

            sort_index = np.argsort(aStream_order)            
            sort_index = sort_index[::-1]
            nUpstream = len(aSegment_upstream)
            for i in sort_index:
                retrieve_flowline_intersect_index(aSegment_upstream[i], aStream_order[i], aVertex_end_upstream[i])
                pass
            pass

        return 

    #starting from the outlet
    dDiatance_min=0.0
    iFlag_first=1
    for j in range(nCell):
        pCell = aCell_intersect_in[j]
        aFlowline= pCell.aFlowline
        nFlowline = len(aFlowline)
        for i in range(nFlowline):
            pFlowline = aFlowline[i]
            pVertex_start = pFlowline.pVertex_start
            pVertex_end = pFlowline.pVertex_end
            dDiatance = pVertex_end.calculate_distance( pVertex_outlet_in)
            if iFlag_first ==1:
                dDiatance_min = dDiatance               
                lCellID_outlet = pCell.lCellID               
                lCellID_outlet2 = j
                lCellID_outlet3 = i
                iFlag_first=0
            else:
                if  dDiatance < dDiatance_min:
                    dDiatance_min = dDiatance                
                    lCellID_outlet = pCell.lCellID                
                    lCellID_outlet2 = j
                    lCellID_outlet3 = i
                    pass    
                else:
                    pass

            pass     

    #loop through       
    pFlowline = aCell_intersect_in[lCellID_outlet2].aFlowline[lCellID_outlet3]
    iStream_segment =pFlowline.iStream_segment
    iStream_order=pFlowline.iStream_order
    pVertex_outlet=pFlowline.pVertex_end
        
    retrieve_flowline_intersect_index(iStream_segment, iStream_order, pVertex_outlet)

    return aFlowline_out, lCellID_outlet, pVertex_outlet




import numpy as np
from pyflowline.algorithms.auxiliary.find_index_in_list import find_vertex_in_list

lID = 0
def merge_flowline(aFlowline_in, aVertex_in, \
    pVertex_outlet_in, \
        aIndex_headwater_in,\
            aIndex_middle_in, \
                aIndex_confluence_in ):

    nVertex=len(aVertex_in)
    nFlowline = len(aFlowline_in)
    aFlowline_out=list()               
    aVertex = np.array(aVertex_in)
    aIndex_headwater = np.array(aIndex_headwater_in)
    aIndex_middle = np.array(aIndex_middle_in)
    aIndex_confluence = np.array(aIndex_confluence_in)
    if aIndex_middle.size == 0:
        return aFlowline_in
    
    aVertex_headwater=aVertex[aIndex_headwater]    
    aVertex_middle=aVertex[aIndex_middle]   
    if aIndex_confluence.size > 0:        
        aVertex_confluence=aVertex[aIndex_confluence]    
    
    def merge_flowline_reach(lIndex_in, pVertex_start_in, pVertex_end_in):
        global lID
        pFlowline = aFlowline_in[lIndex_in]
        pVertex_current = pVertex_start_in
        dummy = lIndex_in
        while (find_vertex_in_list(aVertex_middle, pVertex_current)[0] ==1):            
            for j in range(0, nFlowline):      
                pFlowline2 = aFlowline_in[j]                
                pVertex_start = pFlowline2.pVertex_start
                pVertex_end = pFlowline2.pVertex_end
                if pVertex_end == pVertex_current:
                    pFlowline = pFlowline.merge_upstream(pFlowline2)
                    pVertex_current = pVertex_start
                    dummy = j
                    break
                else:
                    pass

        pFlowline.lIndex = lID
        aFlowline_out.append(pFlowline)        
        lID = lID + 1        
        #go to next 
        if find_vertex_in_list(aVertex_headwater, pVertex_current)[0] ==1: 
            return
        else:
            #confluence
            if find_vertex_in_list(aVertex_confluence, pVertex_current)[0] ==1: 
                for k in range(0, nFlowline):                      
                    pFlowline3 = aFlowline_in[k]                
                    pVertex_start = pFlowline3.pVertex_start
                    pVertex_end = pFlowline3.pVertex_end
                    if pVertex_end == pVertex_current:
                        merge_flowline_reach(k, pVertex_start, pVertex_end)
                                
    
    iFlag_first=1
    for i in range(nFlowline):        
        pFlowline = aFlowline_in[i]                
        pVertex_start = pFlowline.pVertex_start
        pVertex_end = pFlowline.pVertex_end
        dDiatance = pVertex_end.calculate_distance( pVertex_outlet_in)
        if iFlag_first ==1:
            dDiatance_min = dDiatance                
            lIndex_outlet = i            
            iFlag_first=0    
        else:            
            if  dDiatance < dDiatance_min:
                dDiatance_min = dDiatance           
                lIndex_outlet = i                
                pass    
            else:
                pass
            
    pFlowline = aFlowline_in[lIndex_outlet]            
    pVertex_start = pFlowline.pVertex_start
    pVertex_end = pFlowline.pVertex_end     
    merge_flowline_reach(lIndex_outlet, pVertex_start, pVertex_end)   
    return aFlowline_out




import numpy as np
from pyflowline.classes.vertex import pyvertex
from pyflowline.classes.edge import pyedge
from pyflowline.classes.cell import pycell

class pytin(pycell):    
    nFlowline=0
    nVertex =0 
    nEdge=0
    dLength=0.0
    dArea=0.0
    dX_center_meter=0.0
    dY_center_meter=0.0
    aEdge=None
    aVertex=None
    aFlowline=None
    lCellID  = -1
    aNeighbor=None #the global ID of all neighbors
    nNeighbor=-1
    def __init__(self, aEdge,aVertex, dLon, dLat):       
        nEdge = len(aEdge)
        if nEdge !=3:
            pass
        else:           
            self.aEdge = aEdge
            self.aVertex = aVertex #the first one and last one are the same
            self.nEdge = len(aEdge)
            self.nVertex = len(aVertex) - 1
            self.dLongitude_center = dLon
            self.dLatitude_center = dLat
            pVertex = dict()        
            pVertex['lon'] =self.dLongitude_center
            pVertex['lat'] =self.dLatitude_center           
            self.pVertex_center = pyvertex(pVertex)
            pass
        pass    
   
    def has_this_edge(self, pEdge_in):
        iFlag_found = 0
        for pEdge in self.aEdge:
            if pEdge.is_overlap(pEdge_in):
                iFlag_found =1 
                break
            else:
                pass       
        
        return iFlag_found

    def which_edge_cross_this_vertex(self, pVertex_in):
        iFlag_found = 0
        pEdge_out = None
        for pEdge in self.aEdge:
            iFlag, dummy ,diff = pEdge.check_vertex_on_edge(pVertex_in)
            if( iFlag ==1 ):
                iFlag_found =1
                pEdge_out = pEdge
                break
            else:
                pass

        return iFlag_found, pEdge_out
    
    def calculate_cell_area(self):           
        self.dArea = 0.0
        return self.dArea

    def calculate_edge_length(self):        
        self.dLength_edge =0.0
        return self.dLength_edge
    
    def share_edge(self, other):
        iFlag_share = 0
        for pEdge in self.aEdge:
            for pEdge2 in other.aEdge:
                if pEdge.is_overlap(pEdge2) ==1 :
                    iFlag_share = 1 
                    break

        return iFlag_share




import numpy as np
from pyflowline.classes.edge import pyedge
from pyflowline.classes.flowline import pyflowline
from pyflowline.algorithms.auxiliary.find_index_in_list import find_vertex_in_list, find_vertex_on_edge
def split_flowline(aFlowline_in, aVertex_in, iFlag_intersect = None, iFlag_use_id=None):
    aFlowline_out = list()
    nFlowline = len(aFlowline_in)
    for i in range(nFlowline):
        pFlowline = aFlowline_in[i]
        iStream_order = pFlowline.iStream_order
        iStream_segment = pFlowline.iStream_segment
        lFlowlineID = pFlowline.lFlowlineID        
        iFlag_dam = pFlowline.iFlag_dam
        nVertex = pFlowline.nVertex
        nEdge= pFlowline.nEdge
        iPart = 0        
        aVertex  = list() #the actual vertex of ROI
        aVertex_all = list() #include vertex that is not ROI, but we need them to subset 
        for j in range(nEdge):
            pEdge=pFlowline.aEdge[j]
            pVertex = pEdge.pVertex_start
            aVertex_all.append(pVertex)
            #get the start first
            iFlag_exist, lIndex = find_vertex_in_list( aVertex_in,  pVertex)
            if iFlag_exist == 1:                
                iPart = iPart + 1
                aVertex.append(pVertex)            
                
            if iFlag_intersect is not None:
                if iFlag_use_id is not None:
                    aDistance = list()
                    iFlag_exist=0
                    aVertex_dummy=list()
                    aIndex=list()
                    npoint =0
                    for k in range(len(aVertex_in)):
                        pVertex = aVertex_in[k]
                        if pVertex.lFlowlineID == lFlowlineID:
                            iFlag_exist =1
                            iPart = iPart + 1                            
                            distance  = pEdge.pVertex_start.calculate_distance(pVertex)
                            aDistance.append(distance)
                            aVertex_dummy.append(pVertex)
                            aIndex.append(k) 
                            npoint= npoint+ 1
                            
                    #sort needed
                    if iFlag_exist == 1 :
                        x = np.array(aDistance)
                        b = np.argsort(x)
                        c = np.array(aIndex)
                        d= c[b]
                        aIndex_order = list(d)
                        #then push back
                        for k in range(npoint):
                            pVertex_dummy = aVertex_in[ aIndex_order[k]  ] 
                            aVertex.append(pVertex_dummy)
                            aVertex_all.append(pVertex_dummy)

                else:
                    iFlag_exist, npoint, aIndex = find_vertex_on_edge( aVertex_in,  pEdge)
                    #they must be ordered
                    if iFlag_exist==1:
                        for m in range(npoint):
                            pVertex_dummy = aVertex_in[aIndex[m]]
                            iPart = iPart + 1
                            aVertex.append(pVertex_dummy)
                            aVertex_all.append(pVertex_dummy)
                        pass

        #the last ending vertex
        pVertex = pFlowline.pVertex_end
        aVertex_all.append(pVertex)
        iFlag_exist, lIndex = find_vertex_in_list( aVertex_in,  pVertex)
        if iFlag_exist == 1:
            iPart = iPart + 1
            aVertex.append(pVertex)
            pass

        if iPart == 0 :
            print('Something is wrong')
            pass
        else:
            if iPart ==1:
                #print('This flowline does not form any loop')
                if iFlag_use_id is not None:
                    pass
                pass
            else:
                if iPart >=2:
                    nLine = iPart-1
                    #rebuild index
                    aVertex_index =list()
                    for m in range(iPart):
                        pVertex= aVertex[m]
                        iFlag_exist, lIndex = find_vertex_in_list( aVertex_all,  pVertex)
                        if iFlag_exist ==1:
                            aVertex_index.append(lIndex)
                    
                    #find duplicate
                    for k in range(nLine):
                        t = aVertex_index[k]
                        s = aVertex_index[k+1]
                        if s!=t:
                            aEdge=list()
                            for l in range(t,s):
                                pVertex0 = aVertex_all[l]  
                                pVertex1 = aVertex_all[l+1]  
                                pEdge = pyedge(pVertex0, pVertex1)
                                aEdge.append(pEdge)
                                pass

                            pFlowline1 = pyflowline(aEdge)
                            pFlowline1.iStream_order = iStream_order
                            pFlowline1.iFlag_dam = iFlag_dam
                            aFlowline_out.append(pFlowline1)
                        pass                    
                    pass
                pass   

    return aFlowline_out





def define_stream_segment_index(aFlowline_in):
    nFlowline = len(aFlowline_in)
    aFlowline_out = list()
    aStream_segment = list()
    if nFlowline == 0 :
        print ('data incomplete')
    else:  
        for i in range(nFlowline):
            pFlowline = aFlowline_in[i]
            pFlowline.iStream_segment = nFlowline - i  
            aStream_segment.append(pFlowline.iStream_segment)
            aFlowline_out.append(pFlowline)

        pass
    return aFlowline_out, aStream_segment